#!/usr/bin/env python3

# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd

from typing import Optional
from typing import List
from typing import Dict
from typing import Tuple
from typing import Any


def check_input(
    parameter: Any,
    parameter_name: str,
    supported_class: Any,
    supported_subclass: Optional[Any] = None,
) -> bool:
    r"""Checks if the given parameter is of the specified type.

    Function that checks if a given parameter is of the specified type and if iterable, all elements are of the specified element type.
    This is mostly an input check function to catch any errors arising from not supported inputs early.

    Parameters
    ----------
    parameter : any
        Parameter to check class of.
    parameter_name : str
        Name of the parameter.
    supported_class : any
        Class the parameter has to be of.
    supported_subclass : any, or None, default = None
        Class of the values in case the parameter is a list or dict.

    Returns
    -------
    bool
        If the given input is okay.

    Raises
    ------
    TypeError
        If the parameter is not of the given class.

    Examples
    --------
    >>> from pyXLMS.data import check_input
    >>> check_input("PEPTIDE", "peptide_a", str)
    True

    >>> from pyXLMS.data import check_input
    >>> check_input([1, 2], "xl_position_proteins_a", list, int)
    True
    """
    if not isinstance(parameter, supported_class):
        raise TypeError(f"{parameter_name} must be {supported_class}!")
    if isinstance(parameter, list) and supported_subclass is not None:
        for value in parameter:
            if not isinstance(value, supported_subclass):
                raise TypeError(
                    f"List values of {parameter_name} must be {supported_subclass}!"
                )
    if isinstance(parameter, dict) and supported_subclass is not None:
        for key in parameter:
            if not isinstance(parameter[key], supported_subclass):
                raise TypeError(
                    f"Dict values of {parameter_name} must be {supported_subclass}!"
                )
    return True


def check_input_multi(
    parameter: Any,
    parameter_name: str,
    supported_classes: List[Any],
    supported_subclass: Optional[Any] = None,
) -> bool:
    r"""Checks if the given parameter is of one of the specified types.

    Function that checks if a given parameter is of one of the specified types and if iterable, all elements are of the specified element type.
    This is mostly an input check function to catch any errors arising from not supported inputs early.

    Parameters
    ----------
    parameter : any
        Parameter to check class of.
    parameter_name : str
        Name of the parameter.
    supported_class : list of any
        Classes the parameter has to be of.
    supported_subclass : any, or None, default = None
        Class of the values in case the parameter is a list or dict.

    Returns
    -------
    bool
        If the given input is okay.

    Raises
    ------
    TypeError
        If the parameter is not of one of the given classes.

    Examples
    --------
    >>> from pyXLMS.data import check_input_multi
    >>> check_input_multi("PEPTIDE", "peptide_a", [str, list])
    True
    """
    if not isinstance(parameter, tuple(supported_classes)):
        raise TypeError(
            f"{parameter_name} must be one of {','.join([str(c) for c in supported_classes])}!"
        )
    if isinstance(parameter, list) and supported_subclass is not None:
        for value in parameter:
            if not isinstance(value, supported_subclass):
                raise TypeError(
                    f"List values of {parameter_name} must be {supported_subclass}!"
                )
    if isinstance(parameter, dict) and supported_subclass is not None:
        for key in parameter:
            if not isinstance(parameter[key], supported_subclass):
                raise TypeError(
                    f"Dict values of {parameter_name} must be {supported_subclass}!"
                )
    return True


def check_indexing(value: int | List[int]) -> bool:
    r"""Checks that the given value is not 0-based.

    Parameters
    ----------
    value : int, or list of int
        The value(s) to check.

    Returns
    -------
    bool
        If the given value(s) is/are okay.

    Raises
    ------
    ValueError
        If any of the values are smaller than one.

    Examples
    --------
    >>> from pyXLMS.data import check_indexing
    >>> check_indexing([1, 2, 3])
    True
    """
    check_input_multi(value, "value", [int, list], int)
    if isinstance(value, int):
        if value < 1:
            raise ValueError(
                "0-based value found! All positions must use 1-based indexing!"
            )
    else:
        for val in value:
            if val < 1:
                raise ValueError(
                    "0-based value found! All positions must use 1-based indexing!"
                )
    return True


def create_crosslink(
    peptide_a: str,
    xl_position_peptide_a: int,
    proteins_a: Optional[List[str]],
    xl_position_proteins_a: Optional[List[int]],
    decoy_a: Optional[bool],
    peptide_b: str,
    xl_position_peptide_b: int,
    proteins_b: Optional[List[str]],
    xl_position_proteins_b: Optional[List[int]],
    decoy_b: Optional[bool],
    score: Optional[float],
    additional_information: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    r"""Creates a crosslink data structure.

    Contains minimal data necessary for representing a single crosslink. The returned crosslink data structure is a dictionary with keys
    as detailed in the return section.

    Parameters
    ----------
    peptide_a : str
        The unmodified amino acid sequence of the first peptide.
    xl_position_peptide_a : int
        The position of the crosslinker in the sequence of the first peptide (1-based).
    proteins_a : list of str, or None
        The accessions of proteins that the first peptide is associated with.
    xl_position_proteins_a : list of int, or None
        Positions of the crosslink in the proteins of the first peptide (1-based).
    decoy_a : bool, or None
        Whether the alpha peptide is from the decoy database or not.
    peptide_b : str
        The unmodified amino acid sequence of the second peptide.
    xl_position_peptide_b : int
        The position of the crosslinker in the sequence of the second peptide (1-based).
    proteins_b : list of str, or None
        The accessions of proteins that the second peptide is associated with.
    xl_position_proteins_b : list of int, or None
        Positions of the crosslink in the proteins of the second peptide (1-based).
    decoy_b : bool, or None
        Whether the beta peptide is from the decoy database or not.
    score: float, or None
        Score of the crosslink.
    additional_information: dict with str keys, or None, default = None
        A dictionary with additional information associated with the crosslink.

    Returns
    -------
    dict
        The dictionary representing the crosslink with keys ``data_type``, ``completeness``, ``alpha_peptide``, ``alpha_peptide_crosslink_position``,
        ``alpha_proteins``, ``alpha_proteins_crosslink_positions``, ``alpha_decoy``, ``beta_peptide``, ``beta_peptide_crosslink_position``,
        ``beta_proteins``, ``beta_proteins_crosslink_positions``, ``beta_decoy``, ``crosslink_type``, ``score``, and ``additional_information``.
        Alpha and beta are assigned based on peptide sequence, the peptide that alphabetically comes first is assigned to alpha.

    Raises
    ------
    TypeError
        If the parameter is not of the given class.
    ValueError
        If the length of crosslink positions is not equal to the length of proteins.

    Notes
    -----
    The minimum required data for creating a crosslink is:

    - ``peptide_a``: The unmodified amino acid sequence of the first peptide.
    - ``peptide_b``: The unmodified amino acid sequence of the second peptide.
    - ``xl_position_peptide_a``: The position of the crosslinker in the sequence of the first peptide (1-based).
    - ``xl_position_peptide_b``: The position of the crosslinker in the sequence of the second peptide (1-based).

    Examples
    --------
    >>> from pyXLMS.data import create_crosslink
    >>> minimal_crosslink = create_crosslink(
    ...     peptide_a="PEPTIDEA",
    ...     xl_position_peptide_a=1,
    ...     proteins_a=None,
    ...     xl_position_proteins_a=None,
    ...     decoy_a=None,
    ...     peptide_b="PEPTIDEB",
    ...     xl_position_peptide_b=5,
    ...     proteins_b=None,
    ...     xl_position_proteins_b=None,
    ...     decoy_b=None,
    ...     score=None,
    ... )

    >>> from pyXLMS.data import create_crosslink
    >>> crosslink = create_crosslink(
    ...     peptide_a="PEPTIDEA",
    ...     xl_position_peptide_a=1,
    ...     proteins_a=["PROTEINA"],
    ...     xl_position_proteins_a=[1],
    ...     decoy_a=False,
    ...     peptide_b="PEPTIDEB",
    ...     xl_position_peptide_b=5,
    ...     proteins_b=["PROTEINB"],
    ...     xl_position_proteins_b=[3],
    ...     decoy_b=False,
    ...     score=34.5,
    ... )
    """
    ## input checks
    full = check_input(peptide_a, "peptide_a", str)
    full = check_input(peptide_b, "peptide_b", str)
    full = check_input(xl_position_peptide_a, "xl_position_peptide_a", int)
    full = check_input(xl_position_peptide_b, "xl_position_peptide_b", int)
    full = (
        full and check_input(proteins_a, "proteins_a", list, str)
        if proteins_a is not None
        else False
    )
    full = (
        full and check_input(proteins_b, "proteins_b", list, str)
        if proteins_b is not None
        else False
    )
    full = (
        full
        and check_input(xl_position_proteins_a, "xl_position_proteins_a", list, int)
        if xl_position_proteins_a is not None
        else False
    )
    full = (
        full
        and check_input(xl_position_proteins_b, "xl_position_proteins_b", list, int)
        if xl_position_proteins_b is not None
        else False
    )
    full = (
        full and check_input(decoy_a, "decoy_a", bool) if decoy_a is not None else False
    )
    full = (
        full and check_input(decoy_b, "decoy_b", bool) if decoy_b is not None else False
    )
    full = full and check_input(score, "score", float) if score is not None else False
    if proteins_a is not None and xl_position_proteins_a is not None:
        if len(proteins_a) != len(xl_position_proteins_a):
            raise ValueError(
                "Crosslink position has to be given for every protein! Length of proteins_a and xl_position_proteins_a has to match!"
            )
    if proteins_b is not None and xl_position_proteins_b is not None:
        if len(proteins_b) != len(xl_position_proteins_b):
            raise ValueError(
                "Crosslink position has to be given for every protein! Length of proteins_b and xl_position_proteins_b has to match!"
            )
    _ok = check_indexing(xl_position_peptide_a)
    _ok = check_indexing(xl_position_peptide_b)
    _ok = (
        check_indexing(xl_position_proteins_a)
        if xl_position_proteins_a is not None
        else True
    )
    _ok = (
        check_indexing(xl_position_proteins_b)
        if xl_position_proteins_b is not None
        else True
    )
    ## processing
    key_a = f"{peptide_a.strip()}{xl_position_peptide_a}"
    key_b = f"{peptide_b.strip()}{xl_position_peptide_b}"
    # if homomeric crosslink
    if key_a == key_b:
        key_a += "_0"
        key_b += "_1"
    crosslink = {
        key_a: {
            "peptide": peptide_a,
            "xl_position_peptide": xl_position_peptide_a,
            "proteins": proteins_a,
            "xl_position_proteins": xl_position_proteins_a,
            "decoy": decoy_a,
        },
        key_b: {
            "peptide": peptide_b,
            "xl_position_peptide": xl_position_peptide_b,
            "proteins": proteins_b,
            "xl_position_proteins": xl_position_proteins_b,
            "decoy": decoy_b,
        },
    }
    keys = sorted(list(crosslink.keys()))
    alpha_proteins = (
        [protein.strip() for protein in crosslink[keys[0]]["proteins"]]
        if crosslink[keys[0]]["proteins"] is not None
        else []
    )
    beta_proteins = (
        [protein.strip() for protein in crosslink[keys[1]]["proteins"]]
        if crosslink[keys[1]]["proteins"] is not None
        else []
    )
    return {
        "data_type": "crosslink",
        "completeness": "full" if full else "partial",
        "alpha_peptide": crosslink[keys[0]]["peptide"].strip(),
        "alpha_peptide_crosslink_position": crosslink[keys[0]]["xl_position_peptide"],
        "alpha_proteins": alpha_proteins if len(alpha_proteins) > 0 else None,
        "alpha_proteins_crosslink_positions": crosslink[keys[0]][
            "xl_position_proteins"
        ],
        "alpha_decoy": crosslink[keys[0]]["decoy"],
        "beta_peptide": crosslink[keys[1]]["peptide"].strip(),
        "beta_peptide_crosslink_position": crosslink[keys[1]]["xl_position_peptide"],
        "beta_proteins": beta_proteins if len(beta_proteins) > 0 else None,
        "beta_proteins_crosslink_positions": crosslink[keys[1]]["xl_position_proteins"],
        "beta_decoy": crosslink[keys[1]]["decoy"],
        "crosslink_type": "intra"
        if len(set(alpha_proteins).intersection(set(beta_proteins))) > 0
        else "inter",
        "score": score if not pd.isna(score) else None,  # pyright: ignore[reportGeneralTypeIssues]
        "additional_information": additional_information,
    }


def create_crosslink_min(
    peptide_a: str,
    xl_position_peptide_a: int,
    peptide_b: str,
    xl_position_peptide_b: int,
    **kwargs,
) -> Dict[str, Any]:
    r"""Creates a crosslink data structure from minimal input.

    Contains minimal data necessary for representing a single crosslink. This is an alias for
    ``data.create_crosslink()``that sets all optional parameters to ``None`` for convenience.
    The returned crosslink data structure is a dictionary with keys as detailed in the return
    section.

    Parameters
    ----------
    peptide_a : str
        The unmodified amino acid sequence of the first peptide.
    xl_position_peptide_a : int
        The position of the crosslinker in the sequence of the first peptide (1-based).
    peptide_b : str
        The unmodified amino acid sequence of the second peptide.
    xl_position_peptide_b : int
        The position of the crosslinker in the sequence of the second peptide (1-based).
    **kwargs
        Any additional parameters will be passed to ``data.create_crosslink()``.

    Returns
    -------
    dict
        The dictionary representing the crosslink with keys ``data_type``, ``completeness``, ``alpha_peptide``, ``alpha_peptide_crosslink_position``,
        ``alpha_proteins``, ``alpha_proteins_crosslink_positions``, ``alpha_decoy``, ``beta_peptide``, ``beta_peptide_crosslink_position``,
        ``beta_proteins``, ``beta_proteins_crosslink_positions``, ``beta_decoy``, ``crosslink_type``, ``score``, and ``additional_information``.
        Alpha and beta are assigned based on peptide sequence, the peptide that alphabetically comes first is assigned to alpha.

    Notes
    -----
    See also ``data.create_crosslink()``.

    Examples
    --------
    >>> from pyXLMS.data import create_crosslink_min
    >>> minimal_crosslink = create_crosslink_min("PEPTIDEA", 1, "PEPTIDEB", 5)
    """
    return create_crosslink(
        peptide_a=peptide_a,
        xl_position_peptide_a=xl_position_peptide_a,
        proteins_a=kwargs["proteins_a"] if "proteins_a" in kwargs else None,
        xl_position_proteins_a=kwargs["xl_position_proteins_a"]
        if "xl_position_proteins_a" in kwargs
        else None,
        decoy_a=kwargs["decoy_a"] if "decoy_a" in kwargs else None,
        peptide_b=peptide_b,
        xl_position_peptide_b=xl_position_peptide_b,
        proteins_b=kwargs["proteins_b"] if "proteins_b" in kwargs else None,
        xl_position_proteins_b=kwargs["xl_position_proteins_b"]
        if "xl_position_proteins_b" in kwargs
        else None,
        decoy_b=kwargs["decoy_b"] if "decoy_b" in kwargs else None,
        score=kwargs["score"] if "score" in kwargs else None,
        additional_information=kwargs["additional_information"]
        if "additional_information" in kwargs
        else None,
    )


def create_crosslink_from_csm(csm: Dict[str, Any]) -> Dict[str, Any]:
    r"""Creates a crosslink data structure from a crosslink-spectrum-match.

    Creates a crosslink data structure from a crosslink-spectrum-match. The returned crosslink data structure is a dictionary with keys
    as detailed in the return section.

    Parameters
    ----------
    csm : dict of str
        The crosslink-spectrum-match item to be converted to a crosslink item.

    Returns
    -------
    dict
        The dictionary representing the crosslink with keys ``data_type``, ``completeness``, ``alpha_peptide``, ``alpha_peptide_crosslink_position``,
        ``alpha_proteins``, ``alpha_proteins_crosslink_positions``, ``alpha_decoy``, ``beta_peptide``, ``beta_peptide_crosslink_position``,
        ``beta_proteins``, ``beta_proteins_crosslink_positions``, ``beta_decoy``, ``crosslink_type``, ``score``, and ``additional_information``.
        Alpha and beta are assigned based on peptide sequence, the peptide that alphabetically comes first is assigned to alpha.

    Raises
    ------
    TypeError
        If parameter ``csm`` is not a valid crosslink-spectrum-match.

    Notes
    -----
    See also ``data.create_crosslink()``.

    Examples
    --------
    >>> from pyXLMS.data import create_csm_min, create_crosslink_from_csm
    >>> csm = create_csm_min("PEPTIDEA", 1, "PEPTIDEB", 5, "RUN_1", 1)
    >>> crosslink = create_crosslink_from_csm(csm)
    """
    _ok = check_input(csm, "csm", dict)
    if "data_type" not in csm or csm["data_type"] != "crosslink-spectrum-match":
        raise TypeError("Parameter csm is not a valid crosslink-spectrum-match!")
    return create_crosslink(
        peptide_a=csm["alpha_peptide"],
        xl_position_peptide_a=csm["alpha_peptide_crosslink_position"],
        proteins_a=csm["alpha_proteins"],
        xl_position_proteins_a=csm["alpha_proteins_crosslink_positions"],
        decoy_a=csm["alpha_decoy"],
        peptide_b=csm["beta_peptide"],
        xl_position_peptide_b=csm["beta_peptide_crosslink_position"],
        proteins_b=csm["beta_proteins"],
        xl_position_proteins_b=csm["beta_proteins_crosslink_positions"],
        decoy_b=csm["beta_decoy"],
        score=csm["score"],
        additional_information=csm["additional_information"],
    )


def create_csm(
    peptide_a: str,
    modifications_a: Optional[Dict[int, Tuple[str, float]]],
    xl_position_peptide_a: int,
    proteins_a: Optional[List[str]],
    xl_position_proteins_a: Optional[List[int]],
    pep_position_proteins_a: Optional[List[int]],
    score_a: Optional[float],
    decoy_a: Optional[bool],
    peptide_b: str,
    modifications_b: Optional[Dict[int, Tuple[str, float]]],
    xl_position_peptide_b: int,
    proteins_b: Optional[List[str]],
    xl_position_proteins_b: Optional[List[int]],
    pep_position_proteins_b: Optional[List[int]],
    score_b: Optional[float],
    decoy_b: Optional[bool],
    score: Optional[float],
    spectrum_file: str,
    scan_nr: int,
    charge: Optional[int],
    rt: Optional[float],
    im_cv: Optional[float],
    additional_information: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    r"""Creates a crosslink-spectrum-match data structure.

    Contains minimal data necessary for representing a single crosslink-spectrum-match. The returned crosslink-spectrum-match data structure
    is a dictionary with keys as detailed in the return section.

    Parameters
    ----------
    peptide_a : str
        The unmodified amino acid sequence of the first peptide.
    modifications_a : dict of [int, tuple], or None
        The modifications of the first peptide given as a dictionary that maps peptide position (1-based) to modification given as a tuple of modification name and modification delta mass.
        ``N-terminal`` modifications should be denoted with position ``0``. ``C-terminal`` modifications should be denoted with position ``len(peptide) + 1``.
        If the peptide is not modified an empty dictionary should be given.
    xl_position_peptide_a : int
        The position of the crosslinker in the sequence of the first peptide (1-based).
    proteins_a : list of str, or None
        The accessions of proteins that the first peptide is associated with.
    xl_position_proteins_a : list of int, or None
        Positions of the crosslink in the proteins of the first peptide (1-based).
    pep_position_proteins_a : list of int, or None
        Positions of the first peptide in the corresponding proteins (1-based).
    score_a : float, or None
        Identification score of the first peptide.
    decoy_a : bool, or None
        Whether the alpha peptide is from the decoy database or not.
    peptide_b : str
        The unmodified amino acid sequence of the second peptide.
    modifications_b : dict of [int, tuple], or None
        The modifications of the second peptide given as a dictionary that maps peptide position (1-based) to modification given as a tuple of modification name and modification delta mass.
        ``N-terminal`` modifications should be denoted with position ``0``. ``C-terminal`` modifications should be denoted with position ``len(peptide) + 1``.
        If the peptide is not modified an empty dictionary should be given.
    xl_position_peptide_b : int
        The position of the crosslinker in the sequence of the second peptide (1-based).
    proteins_b : list of str, or None
        The accessions of proteins that the second peptide is associated with.
    xl_position_proteins_b : list of int, or None
        Positions of the crosslink in the proteins of the second peptide (1-based).
    pep_position_proteins_b : list of int, or None
        Positions of the second peptide in the corresponding proteins (1-based).
    score_b : float, or None
        Identification score of the second peptide.
    decoy_b : bool, or None
        Whether the beta peptide is from the decoy database or not.
    score: float, or None
        Score of the crosslink-spectrum-match.
    spectrum_file : str
        Name of the spectrum file the crosslink-spectrum-match was identified in.
    scan_nr : int
        The corresponding scan number of the crosslink-spectrum-match.
    charge : int, or None
        The precursor charge of the corresponding mass spectrum of the crosslink-spectrum-match.
    rt : float, or None
        The retention time of the corresponding mass spectrum of the crosslink-spectrum-match in seconds.
    im_cv : float, or None
        The ion mobility or compensation voltage of the corresponding mass spectrum of the crosslink-spectrum-match.
    additional_information: dict with str keys, or None, default = None
        A dictionary with additional information associated with the crosslink-spectrum-match.

    Returns
    -------
    dict
        The dictionary representing the crosslink-spectrum-match with keys ``data_type``, ``completeness``, ``alpha_peptide``, ``alpha_modifications``,
        ``alpha_peptide_crosslink_position``, ``alpha_proteins``, ``alpha_proteins_crosslink_positions``, ``alpha_proteins_peptide_positions``,
        ``alpha_score``, ``alpha_decoy``, ``beta_peptide``, ``beta_modifications``, ``beta_peptide_crosslink_position``, ``beta_proteins``,
        ``beta_proteins_crosslink_positions``, ``beta_proteins_peptide_positions``, ``beta_score``, ``beta_decoy``, ``crosslink_type``, ``score``,
        ``spectrum_file``, ``scan_nr``, ``retention_time``, ``ion_mobility``, and ``additional_information``.
        Alpha and beta are assigned based on peptide sequence, the peptide that alphabetically comes first is assigned to alpha.

    Raises
    ------
    TypeError
        If the parameter is not of the given class.
    ValueError
        If the length of crosslink positions or peptide positions is not equal to the length of proteins.

    Notes
    -----
    The minimum required data for creating a crosslink-spectrum-match is:

    - ``peptide_a``: The unmodified amino acid sequence of the first peptide.
    - ``peptide_b``: The unmodified amino acid sequence of the second peptide.
    - ``xl_position_peptide_a``: The position of the crosslinker in the sequence of the first peptide (1-based).
    - ``xl_position_peptide_b``: The position of the crosslinker in the sequence of the second peptide (1-based).
    - ``spectrum_file``: Name of the spectrum file the crosslink-spectrum-match was identified in.
    - ``scan_nr``: The corresponding scan number of the crosslink-spectrum-match.

    Examples
    --------
    >>> from pyXLMS.data import create_csm
    >>> minimal_csm = create_csm(
    ...     peptide_a="PEPTIDEA",
    ...     modifications_a={},
    ...     xl_position_peptide_a=1,
    ...     proteins_a=None,
    ...     xl_position_proteins_a=None,
    ...     pep_position_proteins_a=None,
    ...     score_a=None,
    ...     decoy_a=None,
    ...     peptide_b="PEPTIDEB",
    ...     modifications_b={},
    ...     xl_position_peptide_b=5,
    ...     proteins_b=None,
    ...     xl_position_proteins_b=None,
    ...     pep_position_proteins_b=None,
    ...     score_b=None,
    ...     decoy_b=None,
    ...     score=None,
    ...     spectrum_file="MS_EXP1",
    ...     scan_nr=1,
    ...     charge=None,
    ...     rt=None,
    ...     im_cv=None,
    ... )

    >>> from pyXLMS.data import create_csm
    >>> csm = create_csm(
    ...     peptide_a="PEPTIDEA",
    ...     modifications_a={1: ("Oxidation", 15.994915)},
    ...     xl_position_peptide_a=1,
    ...     proteins_a=["PROTEINA"],
    ...     xl_position_proteins_a=[1],
    ...     pep_position_proteins_a=[1],
    ...     score_a=20.1,
    ...     decoy_a=False,
    ...     peptide_b="PEPTIDEB",
    ...     modifications_b={},
    ...     xl_position_peptide_b=5,
    ...     proteins_b=["PROTEINB"],
    ...     xl_position_proteins_b=[3],
    ...     pep_position_proteins_b=[1],
    ...     score_b=33.7,
    ...     decoy_b=False,
    ...     score=20.1,
    ...     spectrum_file="MS_EXP1",
    ...     scan_nr=1,
    ...     charge=3,
    ...     rt=13.5,
    ...     im_cv=-50,
    ... )
    """
    ## input checks
    full = check_input(peptide_a, "peptide_a", str)
    full = check_input(peptide_b, "peptide_b", str)
    full = check_input(xl_position_peptide_a, "xl_position_peptide_a", int)
    full = check_input(xl_position_peptide_b, "xl_position_peptide_b", int)
    full = (
        full and check_input(modifications_a, "modifications_a", dict, tuple)
        if modifications_a is not None
        else False
    )
    full = (
        full and check_input(modifications_b, "modifications_b", dict, tuple)
        if modifications_b is not None
        else False
    )
    full = (
        full and check_input(proteins_a, "proteins_a", list, str)
        if proteins_a is not None
        else False
    )
    full = (
        full and check_input(proteins_b, "proteins_b", list, str)
        if proteins_b is not None
        else False
    )
    full = (
        full
        and check_input(xl_position_proteins_a, "xl_position_proteins_a", list, int)
        if xl_position_proteins_a is not None
        else False
    )
    full = (
        full
        and check_input(xl_position_proteins_b, "xl_position_proteins_b", list, int)
        if xl_position_proteins_b is not None
        else False
    )
    full = (
        full
        and check_input(pep_position_proteins_a, "pep_position_proteins_a", list, int)
        if pep_position_proteins_a is not None
        else False
    )
    full = (
        full
        and check_input(pep_position_proteins_b, "pep_position_proteins_b", list, int)
        if pep_position_proteins_b is not None
        else False
    )
    full = (
        full and check_input(score_a, "score_a", float)
        if score_a is not None
        else False
    )
    full = (
        full and check_input(score_b, "score_b", float)
        if score_b is not None
        else False
    )
    full = (
        full and check_input(decoy_a, "decoy_a", bool) if decoy_a is not None else False
    )
    full = (
        full and check_input(decoy_b, "decoy_b", bool) if decoy_b is not None else False
    )
    full = full and check_input(score, "score", float) if score is not None else False
    full = full and check_input(spectrum_file, "spectrum_file", str)
    full = full and check_input(scan_nr, "scan_nr", int)
    full = full and check_input(charge, "charge", int) if charge is not None else False
    full = full and check_input(rt, "rt", float) if rt is not None else False
    full = full and check_input(im_cv, "im_cv", float) if im_cv is not None else False
    if proteins_a is not None and xl_position_proteins_a is not None:
        if len(proteins_a) != len(xl_position_proteins_a):
            raise ValueError(
                "Crosslink position has to be given for every protein! Length of proteins_a and xl_position_proteins_a has to match!"
            )
    if proteins_b is not None and xl_position_proteins_b is not None:
        if len(proteins_b) != len(xl_position_proteins_b):
            raise ValueError(
                "Crosslink position has to be given for every protein! Length of proteins_b and xl_position_proteins_b has to match!"
            )
    if proteins_a is not None and pep_position_proteins_a is not None:
        if len(proteins_a) != len(pep_position_proteins_a):
            raise ValueError(
                "Peptide position has to be given for every protein! Length of proteins_a and pep_position_proteins_a has to match!"
            )
    if proteins_b is not None and pep_position_proteins_b is not None:
        if len(proteins_b) != len(pep_position_proteins_b):
            raise ValueError(
                "Peptide position has to be given for every protein! Length of proteins_b and pep_position_proteins_b has to match!"
            )
    _ok = check_indexing(xl_position_peptide_a)
    _ok = check_indexing(xl_position_peptide_b)
    _ok = (
        check_indexing(xl_position_proteins_a)
        if xl_position_proteins_a is not None
        else True
    )
    _ok = (
        check_indexing(xl_position_proteins_b)
        if xl_position_proteins_b is not None
        else True
    )
    _ok = (
        check_indexing(pep_position_proteins_a)
        if pep_position_proteins_a is not None
        else True
    )
    _ok = (
        check_indexing(pep_position_proteins_b)
        if pep_position_proteins_b is not None
        else True
    )
    ## validity
    if xl_position_proteins_a is not None and pep_position_proteins_a is not None:
        for i in range(len(xl_position_proteins_a)):
            if (
                xl_position_proteins_a[i] - pep_position_proteins_a[i] + 1
                != xl_position_peptide_a
            ):
                _ok = check_indexing(0)
    if xl_position_proteins_b is not None and pep_position_proteins_b is not None:
        for i in range(len(xl_position_proteins_b)):
            if (
                xl_position_proteins_b[i] - pep_position_proteins_b[i] + 1
                != xl_position_peptide_b
            ):
                _ok = check_indexing(0)
    ## processing
    key_a = f"{peptide_a.strip()}{xl_position_peptide_a}"
    key_b = f"{peptide_b.strip()}{xl_position_peptide_b}"
    # if homomeric crosslink
    if key_a == key_b:
        key_a += "_0"
        key_b += "_1"
    crosslink = {
        key_a: {
            "peptide": peptide_a,
            "modifications": {
                int(key): (
                    modifications_a[key][0].strip(),
                    float(modifications_a[key][1]),
                )
                for key in modifications_a.keys()
            }
            if modifications_a is not None
            else None,
            "xl_position_peptide": xl_position_peptide_a,
            "proteins": proteins_a,
            "xl_position_proteins": xl_position_proteins_a,
            "pep_position_proteins": pep_position_proteins_a,
            "score": score_a,
            "decoy": decoy_a,
        },
        key_b: {
            "peptide": peptide_b,
            "modifications": {
                int(key): (
                    modifications_b[key][0].strip(),
                    float(modifications_b[key][1]),
                )
                for key in modifications_b.keys()
            }
            if modifications_b is not None
            else None,
            "xl_position_peptide": xl_position_peptide_b,
            "proteins": proteins_b,
            "xl_position_proteins": xl_position_proteins_b,
            "pep_position_proteins": pep_position_proteins_b,
            "score": score_b,
            "decoy": decoy_b,
        },
    }
    keys = sorted(list(crosslink.keys()))
    alpha_proteins = (
        [protein.strip() for protein in crosslink[keys[0]]["proteins"]]
        if crosslink[keys[0]]["proteins"] is not None
        else []
    )
    beta_proteins = (
        [protein.strip() for protein in crosslink[keys[1]]["proteins"]]
        if crosslink[keys[1]]["proteins"] is not None
        else []
    )
    return {
        "data_type": "crosslink-spectrum-match",
        "completeness": "full" if full else "partial",
        "alpha_peptide": crosslink[keys[0]]["peptide"].strip(),
        "alpha_modifications": crosslink[keys[0]]["modifications"],
        "alpha_peptide_crosslink_position": crosslink[keys[0]]["xl_position_peptide"],
        "alpha_proteins": alpha_proteins if len(alpha_proteins) > 0 else None,
        "alpha_proteins_crosslink_positions": crosslink[keys[0]][
            "xl_position_proteins"
        ],
        "alpha_proteins_peptide_positions": crosslink[keys[0]]["pep_position_proteins"],
        "alpha_score": crosslink[keys[0]]["score"]
        if not pd.isna(crosslink[keys[0]]["score"])
        else None,  # pyright: ignore[reportGeneralTypeIssues]
        "alpha_decoy": crosslink[keys[0]]["decoy"],
        "beta_peptide": crosslink[keys[1]]["peptide"].strip(),
        "beta_modifications": crosslink[keys[1]]["modifications"],
        "beta_peptide_crosslink_position": crosslink[keys[1]]["xl_position_peptide"],
        "beta_proteins": beta_proteins if len(beta_proteins) > 0 else None,
        "beta_proteins_crosslink_positions": crosslink[keys[1]]["xl_position_proteins"],
        "beta_proteins_peptide_positions": crosslink[keys[1]]["pep_position_proteins"],
        "beta_score": crosslink[keys[1]]["score"]
        if not pd.isna(crosslink[keys[1]]["score"])
        else None,  # pyright: ignore[reportGeneralTypeIssues]
        "beta_decoy": crosslink[keys[1]]["decoy"],
        "crosslink_type": "intra"
        if len(set(alpha_proteins).intersection(set(beta_proteins))) > 0
        else "inter",
        "score": score if not pd.isna(score) else None,  # pyright: ignore[reportGeneralTypeIssues]
        "spectrum_file": spectrum_file.strip(),
        "scan_nr": scan_nr,
        "charge": charge,
        "retention_time": rt if not pd.isna(rt) else None,  # pyright: ignore[reportGeneralTypeIssues]
        "ion_mobility": im_cv if not pd.isna(im_cv) else None,  # pyright: ignore[reportGeneralTypeIssues]
        "additional_information": additional_information,
    }


def create_csm_min(
    peptide_a: str,
    xl_position_peptide_a: int,
    peptide_b: str,
    xl_position_peptide_b: int,
    spectrum_file: str,
    scan_nr: int,
    **kwargs,
) -> Dict[str, Any]:
    r"""Creates a crosslink-spectrum-match data structure from minimal input.

    Contains minimal data necessary for representing a single crosslink-spectrum-match. This
    is an alias for ``data.create_csm()``that sets all optional parameters to ``None`` for convenience.
    The returned crosslink-spectrum-match data structure is a dictionary with keys as detailed in the
    return section.

    Parameters
    ----------
    peptide_a : str
        The unmodified amino acid sequence of the first peptide.
    xl_position_peptide_a : int
        The position of the crosslinker in the sequence of the first peptide (1-based).
    peptide_b : str
        The unmodified amino acid sequence of the second peptide.
    xl_position_peptide_b : int
        The position of the crosslinker in the sequence of the second peptide (1-based).
    spectrum_file : str
        Name of the spectrum file the crosslink-spectrum-match was identified in.
    scan_nr : int
        The corresponding scan number of the crosslink-spectrum-match.
    **kwargs
        Any additional parameters will be passed to ``data.create_csm()``.

    Returns
    -------
    dict
        The dictionary representing the crosslink-spectrum-match with keys ``data_type``, ``completeness``, ``alpha_peptide``, ``alpha_modifications``,
        ``alpha_peptide_crosslink_position``, ``alpha_proteins``, ``alpha_proteins_crosslink_positions``, ``alpha_proteins_peptide_positions``,
        ``alpha_score``, ``alpha_decoy``, ``beta_peptide``, ``beta_modifications``, ``beta_peptide_crosslink_position``, ``beta_proteins``,
        ``beta_proteins_crosslink_positions``, ``beta_proteins_peptide_positions``, ``beta_score``, ``beta_decoy``, ``crosslink_type``, ``score``,
        ``spectrum_file``, ``scan_nr``, ``retention_time``, ``ion_mobility``, and ``additional_information``.
        Alpha and beta are assigned based on peptide sequence, the peptide that alphabetically comes first is assigned to alpha.

    Notes
    -----
    See also ``data.create_csm()``.

    Examples
    --------
    >>> from pyXLMS.data import create_csm_min
    >>> minimal_csm = create_csm("PEPTIDEA", 1, "PEPTIDEB", 5, "MS_EXP1", 1)
    """
    return create_csm(
        peptide_a=peptide_a,
        modifications_a=kwargs["modifications_a"]
        if "modifications_a" in kwargs
        else None,
        xl_position_peptide_a=xl_position_peptide_a,
        proteins_a=kwargs["proteins_a"] if "proteins_a" in kwargs else None,
        xl_position_proteins_a=kwargs["xl_position_proteins_a"]
        if "xl_position_proteins_a" in kwargs
        else None,
        pep_position_proteins_a=kwargs["pep_position_proteins_a"]
        if "pep_position_proteins_a" in kwargs
        else None,
        score_a=kwargs["score_a"] if "score_a" in kwargs else None,
        decoy_a=kwargs["decoy_a"] if "decoy_a" in kwargs else None,
        peptide_b=peptide_b,
        modifications_b=kwargs["modifications_b"]
        if "modifications_b" in kwargs
        else None,
        xl_position_peptide_b=xl_position_peptide_b,
        proteins_b=kwargs["proteins_b"] if "proteins_b" in kwargs else None,
        xl_position_proteins_b=kwargs["xl_position_proteins_b"]
        if "xl_position_proteins_b" in kwargs
        else None,
        pep_position_proteins_b=kwargs["pep_position_proteins_b"]
        if "pep_position_proteins_b" in kwargs
        else None,
        score_b=kwargs["score_b"] if "score_b" in kwargs else None,
        decoy_b=kwargs["decoy_b"] if "decoy_b" in kwargs else None,
        score=kwargs["score"] if "score" in kwargs else None,
        spectrum_file=spectrum_file,
        scan_nr=scan_nr,
        charge=kwargs["charge"] if "charge" in kwargs else None,
        rt=kwargs["rt"] if "rt" in kwargs else None,
        im_cv=kwargs["im_cv"] if "im_cv" in kwargs else None,
        additional_information=kwargs["additional_information"]
        if "additional_information" in kwargs
        else None,
    )


def create_parser_result(
    search_engine: str,
    csms: Optional[List[Dict[str, Any]]],
    crosslinks: Optional[List[Dict[str, Any]]],
) -> Dict[str, Any]:
    r"""Creates a parser result data structure.

    Contains all necessary data elements that should be contained in a result returned by a crosslink search engine result parser.

    Parameters
    ----------
    search_engine : str
        Name of the identifying crosslink search engine.
    csms : list of dict, or None
        List of crosslink-spectrum-matches as created by ``data.create_csm()``.
    crosslinks : list of dict, or None
        List of crosslinks as created by ``data.create_crosslink()``.

    Returns
    -------
    dict
        The parser result data structure which is a dictionary with keys ``data_type``, ``completeness``, ``search_engine``, ``crosslink-spectrum-matches`` and
        ``crosslinks``.

    Examples
    --------
    >>> from pyXLMS.data import create_parser_result
    >>> result = create_parser_result("MS Annika", None, None)
    >>> result["data_type"]
    'parser_result'
    >>> result["completeness"]
    'empty'
    >>> result["search_engine"]
    'MS Annika'
    """
    completeness = "partial"
    if csms is None and crosslinks is None:
        completeness = "empty"
    if csms is not None and crosslinks is not None:
        completeness = "full"
    return {
        "data_type": "parser_result",
        "completeness": completeness,
        "search_engine": search_engine,
        "crosslink-spectrum-matches": csms,
        "crosslinks": crosslinks,
    }



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

from ..data import check_input
from ..data import check_input_multi
from .aggregate import unique
from .filter import filter_crosslink_type
from .filter import filter_target_decoy

from typing import Dict
from typing import List
from typing import Any


def __summary_csm(data: List[Dict[str, Any]]) -> Dict[str, float]:
    r"""Extracts summary stats from a list of crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslink-spectrum-matches.

    Returns
    -------
    dict of str, float
        A dictionary with summary stats of the list of crosslink-spectrum-matches.

    Notes
    -----
    This function should not be called directly, it is called from ``summary()``.
    """
    # number of CSMs
    nr = float(len(data))
    # number of unique CSMs
    nr_unique = float(len(unique(data)))
    csm_types = filter_crosslink_type(data)
    # number of intra CSMs
    nr_intra = float(len(csm_types["Intra"]))
    # number of inter CSMs
    nr_inter = float(len(csm_types["Inter"]))
    target_decoys = filter_target_decoy(data)
    # number of TT CSMs
    nr_tt = float(len(target_decoys["Target-Target"]))
    # nr of TD CSMs
    nr_td = float(len(target_decoys["Target-Decoy"]))
    # nr of DD CSMs
    nr_dd = float(len(target_decoys["Decoy-Decoy"]))
    scores = [csm["score"] for csm in data if csm["score"] is not None]
    # min CSM score
    min_score = float(min(scores)) if len(scores) > 0 else float("nan")
    # max CSM score
    max_score = float(max(scores)) if len(scores) > 0 else float("nan")
    return {
        "Number of CSMs": nr,
        "Number of unique CSMs": nr_unique,
        "Number of intra CSMs": nr_intra,
        "Number of inter CSMs": nr_inter,
        "Number of target-target CSMs": nr_tt,
        "Number of target-decoy CSMs": nr_td,
        "Number of decoy-decoy CSMs": nr_dd,
        "Minimum CSM score": min_score,
        "Maximum CSM score": max_score,
    }


def __summary_xl(data: List[Dict[str, Any]]) -> Dict[str, float]:
    r"""Extracts summary stats from a list of crosslinks.

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslinks.

    Returns
    -------
    dict of str, float
        A dictionary with summary stats of the list of crosslinks.

    Notes
    -----
    This function should not be called directly, it is called from ``summary()``.
    """
    # number of crosslinks
    nr = float(len(data))
    # number of unique crosslinks by peptide
    nr_unique_peptide = float(len(unique(data, by="peptide")))
    # number of unique crosslinks by protein
    nr_unique_protein = float("nan")
    try:
        nr_unique_protein = float(len(unique(data, by="protein")))
    except Exception as _e:
        pass
    xl_types = filter_crosslink_type(data)
    # number of intra crosslinks
    nr_intra = float(len(xl_types["Intra"]))
    # number of inter crosslinks
    nr_inter = float(len(xl_types["Inter"]))
    target_decoys = filter_target_decoy(data)
    # number of TT crosslinks
    nr_tt = float(len(target_decoys["Target-Target"]))
    # nr of TD crosslinks
    nr_td = float(len(target_decoys["Target-Decoy"]))
    # nr of DD crosslinks
    nr_dd = float(len(target_decoys["Decoy-Decoy"]))
    scores = [xl["score"] for xl in data if xl["score"] is not None]
    # min crosslink score
    min_score = float(min(scores)) if len(scores) > 0 else float("nan")
    # max crosslink score
    max_score = float(max(scores)) if len(scores) > 0 else float("nan")
    return {
        "Number of crosslinks": nr,
        "Number of unique crosslinks by peptide": nr_unique_peptide,
        "Number of unique crosslinks by protein": nr_unique_protein,
        "Number of intra crosslinks": nr_intra,
        "Number of inter crosslinks": nr_inter,
        "Number of target-target crosslinks": nr_tt,
        "Number of target-decoy crosslinks": nr_td,
        "Number of decoy-decoy crosslinks": nr_dd,
        "Minimum crosslink score": min_score,
        "Maximum crosslink score": max_score,
    }


def summary(data: List[Dict[str, Any]] | Dict[str, Any]) -> Dict[str, float]:
    r"""Extracts summary stats from a list of crosslinks or crosslink-spectrum-matches, or a
    parser_result.

    Extracts summary statistics from a list of crosslinks or crosslink-spectrum-matches, or a parser_result.
    The statistic depend on the supplied data type, if a list of crosslinks is supplied a dictionary with the
    following statistics and keys is returned:

    - Number of crosslinks
    - Number of unique crosslinks by peptide
    - Number of unique crosslinks by protein
    - Number of intra crosslinks
    - Number of inter crosslinks
    - Number of target-target crosslinks
    - Number of target-decoy crosslinks
    - Number of decoy-decoy crosslinks
    - Minimum crosslink score
    - Maximum crosslink score

    If a list of crosslink-spectrum-matches is supplied dictionary with the following statistics and keys is
    returned:

    - Number of CSMs
    - Number of unique CSMs
    - Number of intra CSMs
    - Number of inter CSMs
    - Number of target-target CSMs
    - Number of target-decoy CSMs
    - Number of decoy-decoy CSMs
    - Minimum CSM score
    - Maximum CSM score

    If a parser_result is supplied, a dictionary with both containing all of these is returned - if they are available.
    A parser_result that only contains crosslinks will only yield a dictionary with crosslink statistics, and vice versa
    a parser_result that only contains crosslink-spectrum-matches will only yield a dictionary with crosslink-spectrum-
    match statistics. If the parser_result result contains both, then both dictionaries will be merged and returned.
    Please note that in this case a single dictionary is returned, that contains both the keys for crosslinks and
    crosslink-spectrum-matches.

    Statistics are also printed to ``stdout``.

    Parameters
    ----------
    data : list of dict of str, any, or dict of str, any
        A list of crosslinks or crosslink-spectrum-matches, or a parser_result.

    Returns
    -------
    dict of str, float
        A dictionary with summary statistics.

    Examples
    --------
    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import summary
    >>> pr = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> csms = pr["crosslink-spectrum-matches"]
    >>> stats = summary(csms)
    Number of CSMs: 826.0
    Number of unique CSMs: 826.0
    Number of intra CSMs: 803.0
    Number of inter CSMs: 23.0
    Number of target-target CSMs: 786.0
    Number of target-decoy CSMs: 39.0
    Number of decoy-decoy CSMs: 1.0
    Minimum CSM score: 1.11
    Maximum CSM score: 452.99

    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import summary
    >>> pr = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> stats = summary(pr)
    Number of crosslinks: 300.0
    Number of unique crosslinks by peptide: 300.0
    Number of unique crosslinks by protein: 298.0
    Number of intra crosslinks: 279.0
    Number of inter crosslinks: 21.0
    Number of target-target crosslinks: 265.0
    Number of target-decoy crosslinks: 0.0
    Number of decoy-decoy crosslinks: 35.0
    Minimum crosslink score: 1.11
    Maximum crosslink score: 452.99
    """
    _ok = check_input_multi(data, "data", [dict, list])
    if isinstance(data, list):
        _ok = check_input(data, "data", list, dict)
        if "data_type" not in data[0] or data[0]["data_type"] not in [
            "crosslink",
            "crosslink-spectrum-match",
        ]:
            raise TypeError(
                "Unsupported data type for input data! Parameter data has to be a list of crosslink or crosslink-spectrum-match, "
                "or a parser_result!"
            )
        if data[0]["data_type"] == "crosslink-spectrum-match":
            csm_summary = __summary_csm(data)
            for k, v in csm_summary.items():
                print(f"{k}: {v}")
            return csm_summary
        xl_summary = __summary_xl(data)
        for k, v in xl_summary.items():
            print(f"{k}: {v}")
        return xl_summary
    if "data_type" not in data or data["data_type"] != "parser_result":
        raise TypeError(
            "Can't annotate positions for dict. Dict has to be a valid 'parser_result'!"
        )
    csm_summary = (
        __summary_csm(data["crosslink-spectrum-matches"])
        if data["crosslink-spectrum-matches"] is not None
        else {}
    )
    for k, v in csm_summary.items():
        print(f"{k}: {v}")
    xl_summary = (
        __summary_xl(data["crosslinks"]) if data["crosslinks"] is not None else {}
    )
    for k, v in xl_summary.items():
        print(f"{k}: {v}")
    return {**csm_summary, **xl_summary}



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS import data

    csm = data.create_csm(
        "PEPTIDE",
        {1: (" Oxidation ", 15.994915)},
        1,
        ["PROTEIN"],
        [1],
        [1],
        50.3,
        False,
        "EDITPEP",
        {2: ("Oxidation", 15.994915)},
        3,
        ["NIETORP", "PROTEIN"],
        [7, 4],
        [5, 2],
        170.3,
        False,
        170.3,
        "RUN_1",
        1,
        3,
        23.4,
        -50.0,
    )
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "full"
    # alpha
    assert csm["alpha_peptide"] == "EDITPEP"
    assert len(csm["alpha_modifications"]) == 1
    assert 2 in csm["alpha_modifications"]
    assert csm["alpha_modifications"][2][0] == "Oxidation"
    assert csm["alpha_modifications"][2][1] == pytest.approx(15.994915)
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert len(csm["alpha_proteins"]) == 2
    assert csm["alpha_proteins"][0] == "NIETORP"
    assert len(csm["alpha_proteins_crosslink_positions"]) == 2
    assert csm["alpha_proteins_crosslink_positions"][0] == 7
    assert len(csm["alpha_proteins_peptide_positions"]) == 2
    assert csm["alpha_proteins_peptide_positions"][0] == 5
    assert csm["alpha_score"] == pytest.approx(170.3)
    # assert csm["alpha_decoy"] == False
    assert not csm["alpha_decoy"]
    # beta
    assert csm["beta_peptide"] == "PEPTIDE"
    assert len(csm["beta_modifications"]) == 1
    assert 1 in csm["beta_modifications"]
    assert csm["beta_modifications"][1][0] == "Oxidation"
    assert csm["beta_modifications"][1][1] == pytest.approx(15.994915)
    assert csm["beta_peptide_crosslink_position"] == 1
    assert len(csm["beta_proteins"]) == 1
    assert csm["beta_proteins"][0] == "PROTEIN"
    assert len(csm["beta_proteins_crosslink_positions"]) == 1
    assert csm["beta_proteins_crosslink_positions"][0] == 1
    assert len(csm["beta_proteins_peptide_positions"]) == 1
    assert csm["beta_proteins_peptide_positions"][0] == 1
    assert csm["beta_score"] == pytest.approx(50.3)
    # assert csm["beta_decoy"] == False
    assert not csm["beta_decoy"]
    # csm
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(170.3)
    assert csm["spectrum_file"] == "RUN_1"
    assert csm["scan_nr"] == 1
    assert csm["charge"] == 3
    assert csm["retention_time"] == pytest.approx(23.4)
    assert csm["ion_mobility"] == pytest.approx(-50.0)


def test2():
    from pyXLMS import data

    csm = data.create_csm(
        "PEPTIDE",
        {1: (" Oxidation ", 15.994915)},
        1,
        ["PROTEIN"],
        [1],
        [1],
        50.3,
        False,
        "EDITPEP",
        {2: ("Oxidation", 15.994915)},
        3,
        ["NIETORP"],
        [7],
        [5],
        170.3,
        False,
        170.3,
        "RUN_1",
        1,
        3,
        23.4,
        -50.0,
    )
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "full"
    # alpha
    assert csm["alpha_peptide"] == "EDITPEP"
    assert len(csm["alpha_modifications"]) == 1
    assert 2 in csm["alpha_modifications"]
    assert csm["alpha_modifications"][2][0] == "Oxidation"
    assert csm["alpha_modifications"][2][1] == pytest.approx(15.994915)
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert len(csm["alpha_proteins"]) == 1
    assert csm["alpha_proteins"][0] == "NIETORP"
    assert len(csm["alpha_proteins_crosslink_positions"]) == 1
    assert csm["alpha_proteins_crosslink_positions"][0] == 7
    assert len(csm["alpha_proteins_peptide_positions"]) == 1
    assert csm["alpha_proteins_peptide_positions"][0] == 5
    assert csm["alpha_score"] == pytest.approx(170.3)
    # assert csm["alpha_decoy"] == False
    assert not csm["alpha_decoy"]
    # beta
    assert csm["beta_peptide"] == "PEPTIDE"
    assert len(csm["beta_modifications"]) == 1
    assert 1 in csm["beta_modifications"]
    assert csm["beta_modifications"][1][0] == "Oxidation"
    assert csm["beta_modifications"][1][1] == pytest.approx(15.994915)
    assert csm["beta_peptide_crosslink_position"] == 1
    assert len(csm["beta_proteins"]) == 1
    assert csm["beta_proteins"][0] == "PROTEIN"
    assert len(csm["beta_proteins_crosslink_positions"]) == 1
    assert csm["beta_proteins_crosslink_positions"][0] == 1
    assert len(csm["beta_proteins_peptide_positions"]) == 1
    assert csm["beta_proteins_peptide_positions"][0] == 1
    assert csm["beta_score"] == pytest.approx(50.3)
    # assert csm["beta_decoy"] == False
    assert not csm["beta_decoy"]
    # csm
    assert csm["crosslink_type"] == "inter"
    assert csm["score"] == pytest.approx(170.3)
    assert csm["spectrum_file"] == "RUN_1"
    assert csm["scan_nr"] == 1
    assert csm["charge"] == 3
    assert csm["retention_time"] == pytest.approx(23.4)
    assert csm["ion_mobility"] == pytest.approx(-50.0)


def test3():
    from pyXLMS import data

    csm = data.create_csm(
        "EDITPEP",
        {2: ("Oxidation", 15.994915)},
        3,
        ["NIETORP", "PROTEIN"],
        [7, 4],
        [5, 2],
        170.3,
        False,
        "PEPTIDE",
        {1: (" Oxidation ", 15.994915)},
        1,
        ["PROTEIN"],
        [1],
        [1],
        50.3,
        False,
        170.3,
        "RUN_1",
        1,
        3,
        23.4,
        -50.0,
    )
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "full"
    # alpha
    assert csm["alpha_peptide"] == "EDITPEP"
    assert len(csm["alpha_modifications"]) == 1
    assert 2 in csm["alpha_modifications"]
    assert csm["alpha_modifications"][2][0] == "Oxidation"
    assert csm["alpha_modifications"][2][1] == pytest.approx(15.994915)
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert len(csm["alpha_proteins"]) == 2
    assert csm["alpha_proteins"][0] == "NIETORP"
    assert len(csm["alpha_proteins_crosslink_positions"]) == 2
    assert csm["alpha_proteins_crosslink_positions"][0] == 7
    assert len(csm["alpha_proteins_peptide_positions"]) == 2
    assert csm["alpha_proteins_peptide_positions"][0] == 5
    assert csm["alpha_score"] == pytest.approx(170.3)
    # assert csm["alpha_decoy"] == False
    assert not csm["alpha_decoy"]
    # beta
    assert csm["beta_peptide"] == "PEPTIDE"
    assert len(csm["beta_modifications"]) == 1
    assert 1 in csm["beta_modifications"]
    assert csm["beta_modifications"][1][0] == "Oxidation"
    assert csm["beta_modifications"][1][1] == pytest.approx(15.994915)
    assert csm["beta_peptide_crosslink_position"] == 1
    assert len(csm["beta_proteins"]) == 1
    assert csm["beta_proteins"][0] == "PROTEIN"
    assert len(csm["beta_proteins_crosslink_positions"]) == 1
    assert csm["beta_proteins_crosslink_positions"][0] == 1
    assert len(csm["beta_proteins_peptide_positions"]) == 1
    assert csm["beta_proteins_peptide_positions"][0] == 1
    assert csm["beta_score"] == pytest.approx(50.3)
    # assert csm["beta_decoy"] == False
    assert not csm["beta_decoy"]
    # csm
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(170.3)
    assert csm["spectrum_file"] == "RUN_1"
    assert csm["scan_nr"] == 1
    assert csm["charge"] == 3
    assert csm["retention_time"] == pytest.approx(23.4)
    assert csm["ion_mobility"] == pytest.approx(-50.0)


def test4():
    from pyXLMS import data

    csm = data.create_csm(
        "PEPTIDE  ",
        {1: ("    Oxidation ", 15.994915)},
        3,
        ["   PROTEIN"],
        [5],
        [3],
        50.3,
        False,
        "   PEPTIDE",
        {2: ("Oxidation", 15.994915)},
        1,
        ["PROTEIN  "],
        [1],
        [1],
        170.3,
        False,
        170.3,
        "RUN_1   ",
        1,
        3,
        23.4,
        -50.0,
    )
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "full"
    # alpha
    assert csm["alpha_peptide"] == "PEPTIDE"
    assert len(csm["alpha_modifications"]) == 1
    assert 2 in csm["alpha_modifications"]
    assert csm["alpha_modifications"][2][0] == "Oxidation"
    assert csm["alpha_modifications"][2][1] == pytest.approx(15.994915)
    assert csm["alpha_peptide_crosslink_position"] == 1
    assert len(csm["alpha_proteins"]) == 1
    assert csm["alpha_proteins"][0] == "PROTEIN"
    assert len(csm["alpha_proteins_crosslink_positions"]) == 1
    assert csm["alpha_proteins_crosslink_positions"][0] == 1
    assert len(csm["alpha_proteins_peptide_positions"]) == 1
    assert csm["alpha_proteins_peptide_positions"][0] == 1
    assert csm["alpha_score"] == pytest.approx(170.3)
    # assert csm["alpha_decoy"] == False
    assert not csm["alpha_decoy"]
    # beta
    assert csm["beta_peptide"] == "PEPTIDE"
    assert len(csm["beta_modifications"]) == 1
    assert 1 in csm["beta_modifications"]
    assert csm["beta_modifications"][1][0] == "Oxidation"
    assert csm["beta_modifications"][1][1] == pytest.approx(15.994915)
    assert csm["beta_peptide_crosslink_position"] == 3
    assert len(csm["beta_proteins"]) == 1
    assert csm["beta_proteins"][0] == "PROTEIN"
    assert len(csm["beta_proteins_crosslink_positions"]) == 1
    assert csm["beta_proteins_crosslink_positions"][0] == 5
    assert len(csm["beta_proteins_peptide_positions"]) == 1
    assert csm["beta_proteins_peptide_positions"][0] == 3
    assert csm["beta_score"] == pytest.approx(50.3)
    # assert csm["beta_decoy"] == False
    assert not csm["beta_decoy"]
    # csm
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(170.3)
    assert csm["spectrum_file"] == "RUN_1"
    assert csm["scan_nr"] == 1
    assert csm["charge"] == 3
    assert csm["retention_time"] == pytest.approx(23.4)
    assert csm["ion_mobility"] == pytest.approx(-50.0)


def test5():
    from pyXLMS import data

    csm = data.create_csm(
        "PEPTIDE",
        None,
        3,
        None,
        None,
        None,
        None,
        None,
        "PEPTIDE",
        None,
        1,
        None,
        None,
        None,
        None,
        None,
        None,
        "RUN_1",
        1,
        None,
        None,
        None,
    )
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    # alpha
    assert csm["alpha_peptide"] == "PEPTIDE"
    assert csm["alpha_modifications"] is None
    assert csm["alpha_peptide_crosslink_position"] == 1
    assert csm["alpha_proteins"] is None
    assert csm["alpha_proteins_crosslink_positions"] is None
    assert csm["alpha_proteins_peptide_positions"] is None
    assert csm["alpha_score"] is None
    assert csm["alpha_decoy"] is None
    # beta
    assert csm["beta_peptide"] == "PEPTIDE"
    assert csm["beta_modifications"] is None
    assert csm["beta_peptide_crosslink_position"] == 3
    assert csm["beta_proteins"] is None
    assert csm["beta_proteins_crosslink_positions"] is None
    assert csm["beta_proteins_peptide_positions"] is None
    assert csm["beta_score"] is None
    assert csm["beta_decoy"] is None
    # csm
    assert csm["crosslink_type"] == "inter"
    assert csm["score"] is None
    assert csm["spectrum_file"] == "RUN_1"
    assert csm["scan_nr"] == 1
    assert csm["charge"] is None
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None


def test6():
    from pyXLMS import data

    with pytest.raises(TypeError, match=f"modifications_a must be {dict}!"):
        _csm = data.create_csm(
            "PEPTIDE",
            (" Oxidation ", 15.994915),
            1,
            ["PROTEIN"],
            [1],
            [1],
            50.3,
            False,
            "EDITPEP",
            {2: ("Oxidation", 15.994915)},
            3,
            ["NIETORP", "PROTEIN"],
            [7, 4],
            [5, 2],
            170.3,
            False,
            170.3,
            "RUN_1",
            1,
            3,
            23.4,
            -50.0,
        )


def test7():
    from pyXLMS import data

    with pytest.raises(
        TypeError, match=f"Dict values of modifications_b must be {tuple}!"
    ):
        _csm = data.create_csm(
            "PEPTIDE",
            {1: (" Oxidation ", 15.994915)},
            1,
            ["PROTEIN"],
            [1],
            [1],
            50.3,
            False,
            "EDITPEP",
            {2: {"Oxidation": 15.994915}},
            3,
            ["NIETORP", "PROTEIN"],
            [7, 4],
            [5, 2],
            170.3,
            False,
            170.3,
            "RUN_1",
            1,
            3,
            23.4,
            -50.0,
        )


def test8():
    from pyXLMS import data

    with pytest.raises(TypeError, match=f"xl_position_peptide_b must be {int}!"):
        _csm = data.create_csm(
            "PEPTIDE",
            {1: (" Oxidation ", 15.994915)},
            1,
            ["PROTEIN"],
            [1],
            [1],
            50.3,
            False,
            "EDITPEP",
            {2: ("Oxidation", 15.994915)},
            "3",
            ["NIETORP", "PROTEIN"],
            [7, 4],
            [5, 2],
            170.3,
            False,
            170.3,
            "RUN_1",
            1,
            3,
            23.4,
            -50.0,
        )


def test9():
    from pyXLMS import data

    with pytest.raises(
        TypeError, match=f"List values of xl_position_proteins_b must be {int}!"
    ):
        _csm = data.create_csm(
            "PEPTIDE",
            {1: (" Oxidation ", 15.994915)},
            1,
            ["PROTEIN"],
            [1],
            [1],
            50.3,
            False,
            "EDITPEP",
            {2: ("Oxidation", 15.994915)},
            3,
            ["NIETORP", "PROTEIN"],
            [7, "2"],
            [5, 2],
            170.3,
            False,
            170.3,
            "RUN_1",
            1,
            3,
            23.4,
            -50.0,
        )


def test10():
    from pyXLMS import data

    with pytest.raises(
        ValueError,
        match="Crosslink position has to be given for every protein! Length of proteins_a and xl_position_proteins_a has to match!",
    ):
        _csm = data.create_csm(
            "PEPTIDE",
            {1: (" Oxidation ", 15.994915)},
            1,
            ["PROTEIN"],
            [1, 2],
            [1],
            50.3,
            False,
            "EDITPEP",
            {2: ("Oxidation", 15.994915)},
            3,
            ["NIETORP", "PROTEIN"],
            [7, 4],
            [5, 2],
            170.3,
            False,
            170.3,
            "RUN_1",
            1,
            3,
            23.4,
            -50.0,
        )


def test11():
    from pyXLMS import data

    with pytest.raises(
        ValueError,
        match="Crosslink position has to be given for every protein! Length of proteins_b and xl_position_proteins_b has to match!",
    ):
        _csm = data.create_csm(
            "PEPTIDE",
            {1: (" Oxidation ", 15.994915)},
            1,
            ["PROTEIN"],
            [1],
            [1],
            50.3,
            False,
            "EDITPEP",
            {2: ("Oxidation", 15.994915)},
            3,
            ["NIETORP", "PROTEIN"],
            [7],
            [5, 2],
            170.3,
            False,
            170.3,
            "RUN_1",
            1,
            3,
            23.4,
            -50.0,
        )


def test12():
    from pyXLMS import data

    with pytest.raises(
        ValueError,
        match="Peptide position has to be given for every protein! Length of proteins_a and pep_position_proteins_a has to match!",
    ):
        _csm = data.create_csm(
            "PEPTIDE",
            {1: (" Oxidation ", 15.994915)},
            1,
            ["PROTEIN"],
            [1],
            [],
            50.3,
            False,
            "EDITPEP",
            {2: ("Oxidation", 15.994915)},
            3,
            ["NIETORP", "PROTEIN"],
            [7, 4],
            [5, 2],
            170.3,
            False,
            170.3,
            "RUN_1",
            1,
            3,
            23.4,
            -50.0,
        )


def test13():
    from pyXLMS import data

    with pytest.raises(
        ValueError,
        match="Peptide position has to be given for every protein! Length of proteins_b and pep_position_proteins_b has to match!",
    ):
        _csm = data.create_csm(
            "PEPTIDE",
            {1: (" Oxidation ", 15.994915)},
            1,
            ["PROTEIN"],
            [1],
            [1],
            50.3,
            False,
            "EDITPEP",
            {2: ("Oxidation", 15.994915)},
            3,
            ["NIETORP", "PROTEIN"],
            [7, 4],
            [5],
            170.3,
            False,
            170.3,
            "RUN_1",
            1,
            3,
            23.4,
            -50.0,
        )


def test14():
    from pyXLMS import data

    with pytest.raises(
        ValueError,
        match="0-based value found! All positions must use 1-based indexing!",
    ):
        _csm = data.create_csm(
            "PEPTIDE",
            {1: (" Oxidation ", 15.994915)},
            1,
            ["PROTEIN"],
            [1],
            [1],
            50.3,
            False,
            "EDITPEP",
            {2: ("Oxidation", 15.994915)},
            3,
            ["NIETORP", "PROTEIN"],
            [7, 4],
            [5, 4],
            170.3,
            False,
            170.3,
            "RUN_1",
            1,
            3,
            23.4,
            -50.0,
        )



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com


def test1():
    from pyXLMS import transform

    modifications = {1: ("Oxidation", 15.994915), 5: ("Carbamidomethyl", 57.021464)}
    modifications_str = "(1:[Oxidation|15.994915]);(5:[Carbamidomethyl|57.021464])"
    assert transform.modifications_to_str(modifications) == modifications_str


def test2():
    from pyXLMS import transform

    modifications = {1: ("Oxidation", 15.994915)}
    modifications_str = "(1:[Oxidation|15.994915])"
    assert transform.modifications_to_str(modifications) == modifications_str


def test3():
    from pyXLMS import transform

    modifications = dict()
    modifications_str = ""
    assert transform.modifications_to_str(modifications) == modifications_str


def test4():
    from pyXLMS import transform

    modifications = None
    assert transform.modifications_to_str(modifications) is None


def test5():
    from pyXLMS import transform
    from pyXLMS import data

    data_list = [
        data.create_crosslink_min("PEPK", 4, "PKEP", 2),
        data.create_crosslink_min("KPEP", 1, "PEKP", 3),
    ]
    assert transform.assert_data_type_same(data_list)

    data_list = [
        data.create_crosslink_min("PEPK", 4, "PKEP", 2),
        data.create_csm_min("KPEP", 1, "PEKP", 3, "RUN_1", 1),
    ]
    assert not transform.assert_data_type_same(data_list)


def test6():
    from pyXLMS import transform
    from pyXLMS import data

    data_list = [
        data.create_crosslink_min("PEPK", 4, "PKEP", 2),
        data.create_crosslink_min("KPEP", 1, "PEKP", 3),
    ]
    k = transform.get_available_keys(data_list)

    assert k["data_type"]
    assert k["completeness"]
    assert k["alpha_peptide"]
    assert k["alpha_peptide_crosslink_position"]
    assert not k["alpha_proteins"]
    assert not k["alpha_proteins_crosslink_positions"]
    assert not k["alpha_decoy"]
    assert k["beta_peptide"]
    assert k["beta_peptide_crosslink_position"]
    assert not k["beta_proteins"]
    assert not k["beta_proteins_crosslink_positions"]
    assert not k["beta_decoy"]
    assert k["crosslink_type"]
    assert not k["score"]
    assert not k["additional_information"]


def test7():
    from pyXLMS import transform
    from pyXLMS import data

    data_list = [
        data.create_crosslink_min("PEPK", 4, "PKEP", 2, score=170.3),
        data.create_crosslink_min("KPEP", 1, "PEKP", 3, score=13.5),
    ]
    k = transform.get_available_keys(data_list)

    assert k["data_type"]
    assert k["completeness"]
    assert k["alpha_peptide"]
    assert k["alpha_peptide_crosslink_position"]
    assert not k["alpha_proteins"]
    assert not k["alpha_proteins_crosslink_positions"]
    assert not k["alpha_decoy"]
    assert k["beta_peptide"]
    assert k["beta_peptide_crosslink_position"]
    assert not k["beta_proteins"]
    assert not k["beta_proteins_crosslink_positions"]
    assert not k["beta_decoy"]
    assert k["crosslink_type"]
    assert k["score"]
    assert not k["additional_information"]


def test8():
    from pyXLMS import transform
    from pyXLMS import data

    data_list = [
        data.create_crosslink(
            "PEPK",
            4,
            ["PROT1"],
            [4],
            False,
            "PKEP",
            2,
            ["PROT2"],
            [2],
            False,
            170.3,
            {},
        ),
        data.create_crosslink(
            "KPEP", 1, ["PROT3"], [1], False, "PEKP", 3, ["PROT4"], [3], True, 13.5, {}
        ),
    ]
    k = transform.get_available_keys(data_list)

    assert k["data_type"]
    assert k["completeness"]
    assert k["alpha_peptide"]
    assert k["alpha_peptide_crosslink_position"]
    assert k["alpha_proteins"]
    assert k["alpha_proteins_crosslink_positions"]
    assert k["alpha_decoy"]
    assert k["beta_peptide"]
    assert k["beta_peptide_crosslink_position"]
    assert k["beta_proteins"]
    assert k["beta_proteins_crosslink_positions"]
    assert k["beta_decoy"]
    assert k["crosslink_type"]
    assert k["score"]
    assert k["additional_information"]


def test9():
    from pyXLMS import transform
    from pyXLMS import data

    data_list = [
        data.create_csm_min("PEPK", 4, "PKEP", 2, "RUN_1", 1),
        data.create_csm_min("KPEP", 1, "PEKP", 3, "RUN_1", 2),
    ]
    k = transform.get_available_keys(data_list)

    assert k["data_type"]
    assert k["completeness"]
    assert k["alpha_peptide"]
    assert not k["alpha_modifications"]
    assert k["alpha_peptide_crosslink_position"]
    assert not k["alpha_proteins"]
    assert not k["alpha_proteins_crosslink_positions"]
    assert not k["alpha_proteins_peptide_positions"]
    assert not k["alpha_score"]
    assert not k["alpha_decoy"]
    assert k["beta_peptide"]
    assert not k["beta_modifications"]
    assert k["beta_peptide_crosslink_position"]
    assert not k["beta_proteins"]
    assert not k["beta_proteins_crosslink_positions"]
    assert not k["beta_proteins_peptide_positions"]
    assert not k["beta_score"]
    assert not k["beta_decoy"]
    assert k["crosslink_type"]
    assert not k["score"]
    assert k["spectrum_file"]
    assert k["scan_nr"]
    assert not k["charge"]
    assert not k["retention_time"]
    assert not k["ion_mobility"]
    assert not k["additional_information"]


def test10():
    from pyXLMS import transform
    from pyXLMS import data

    data_list = [
        data.create_csm_min("PEPK", 4, "PKEP", 2, "RUN_1", 1, score=170.3),
        data.create_csm_min("KPEP", 1, "PEKP", 3, "RUN_1", 2, score=13.5),
    ]
    k = transform.get_available_keys(data_list)

    assert k["data_type"]
    assert k["completeness"]
    assert k["alpha_peptide"]
    assert not k["alpha_modifications"]
    assert k["alpha_peptide_crosslink_position"]
    assert not k["alpha_proteins"]
    assert not k["alpha_proteins_crosslink_positions"]
    assert not k["alpha_proteins_peptide_positions"]
    assert not k["alpha_score"]
    assert not k["alpha_decoy"]
    assert k["beta_peptide"]
    assert not k["beta_modifications"]
    assert k["beta_peptide_crosslink_position"]
    assert not k["beta_proteins"]
    assert not k["beta_proteins_crosslink_positions"]
    assert not k["beta_proteins_peptide_positions"]
    assert not k["beta_score"]
    assert not k["beta_decoy"]
    assert k["crosslink_type"]
    assert k["score"]
    assert k["spectrum_file"]
    assert k["scan_nr"]
    assert not k["charge"]
    assert not k["retention_time"]
    assert not k["ion_mobility"]
    assert not k["additional_information"]


def test11():
    from pyXLMS import transform
    from pyXLMS import data

    data_list = [
        data.create_csm(
            "PEPK",
            {},
            4,
            ["PROT1"],
            [4],
            [1],
            170.3,
            False,
            "PKEP",
            {},
            2,
            ["PROT2"],
            [2],
            [1],
            150.4,
            False,
            150.4,
            "RUN_1",
            1,
            4,
            213.123,
            -50.0,
            {},
        ),
        data.create_csm(
            "KPEP",
            {},
            1,
            ["PROT3"],
            [1],
            [1],
            171.3,
            False,
            "PEKP",
            {},
            3,
            ["PROT4"],
            [3],
            [1],
            13.5,
            True,
            13.5,
            "RUN_1",
            2,
            3,
            213.124,
            -70.0,
            {},
        ),
    ]
    k = transform.get_available_keys(data_list)

    assert k["data_type"]
    assert k["completeness"]
    assert k["alpha_peptide"]
    assert k["alpha_modifications"]
    assert k["alpha_peptide_crosslink_position"]
    assert k["alpha_proteins"]
    assert k["alpha_proteins_crosslink_positions"]
    assert k["alpha_proteins_peptide_positions"]
    assert k["alpha_score"]
    assert k["alpha_decoy"]
    assert k["beta_peptide"]
    assert k["beta_modifications"]
    assert k["beta_peptide_crosslink_position"]
    assert k["beta_proteins"]
    assert k["beta_proteins_crosslink_positions"]
    assert k["beta_proteins_peptide_positions"]
    assert k["beta_score"]
    assert k["beta_decoy"]
    assert k["crosslink_type"]
    assert k["score"]
    assert k["spectrum_file"]
    assert k["scan_nr"]
    assert k["charge"]
    assert k["retention_time"]
    assert k["ion_mobility"]
    assert k["additional_information"]



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    targets = targets_only(result["crosslink-spectrum-matches"])
    assert len(targets) == 786


def test2():
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    targets = targets_only(result["crosslinks"])
    len(targets) == 265


def test3():
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only

    result = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    result_targets = targets_only(result)
    assert len(result_targets["crosslink-spectrum-matches"]) == 786
    assert len(result_targets["crosslinks"]) == 265


def test4():
    from pyXLMS.transform import targets_only
    from pyXLMS.data import create_crosslink_min

    xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    crosslinks = [xl1, xl2]
    with pytest.raises(
        RuntimeError,
        match=r"No target matches found! Are you sure your data is labelled\?",
    ):
        _r = targets_only(crosslinks)


def test5():
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    result["crosslink-spectrum-matches"] = []
    with pytest.raises(
        RuntimeError,
        match=r"No target crosslink-spectrum-matches found! Are you sure they are labelled\?",
    ):
        _r = targets_only(result)


def test6():
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    result["crosslinks"] = []
    with pytest.raises(
        RuntimeError,
        match=r"No target crosslinks found! Are you sure they are labelled\?",
    ):
        _r = targets_only(result)



#!/usr/bin/env python3

# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import re
import warnings
from tqdm import tqdm
from Bio.SeqIO.FastaIO import SimpleFastaParser

from ..constants import AMINO_ACIDS_REPLACEMENTS
from ..data import check_input
from ..data import create_csm
from ..data import create_crosslink
from ..data import create_parser_result
from .util import assert_data_type_same

from typing import Optional
from typing import BinaryIO
from typing import Callable
from typing import Dict
from typing import Tuple
from typing import List
from typing import Any


def __generate_all_sequences(sequence: str) -> List[str]:
    r"""Generates all possible amino acid sequences for a given amino acid sequence if it contains placeholder amino acids.

    Generates all possible amino acid sequences for a given amino acid sequence if it contains placeholder amino acids. Has no
    effect on amino acid sequences that do not contain placeholder amino acids.

    Parameters
    ----------
    sequence : str
        The amino acid sequence of a peptide or protein that the generation should be based on.

    Returns
    -------
    list of str
        List of generated amino acid sequences without placeholders. If the original sequence did not contain
        any placeholder amino acids, the list will only contain the original sequence.

    Notes
    -----
    This function should not be called directly, it is called from ``__get_proteins_and_positions()``.
    """
    sequence_needs_generation = False
    for one_letter_code in AMINO_ACIDS_REPLACEMENTS:
        if one_letter_code in sequence:
            sequence_needs_generation = True
            break
    if not sequence_needs_generation:
        return [sequence]
    all_sequences = [sequence]
    for one_letter_code, replacements in AMINO_ACIDS_REPLACEMENTS.items():
        occurences = sequence.count(one_letter_code)
        for i in range(occurences):
            all_sequences = [
                seq.replace(one_letter_code, replacement, 1)
                for seq in all_sequences
                for replacement in replacements
            ]
    return all_sequences


def __get_proteins_and_positions(
    peptide: str, protein_db: Dict[str, str]
) -> Tuple[List[str], List[int]]:
    r"""Retrieve matching proteins and peptide positions for a specific peptide.

    Matches the specified peptide against the given protein database and returns all proteins
    that contain the peptide, as well as the corresponding peptide positions in those proteins.
    Uses 0-based indexing.

    Parameters
    ----------
    peptide : str
        Unmodified peptide sequence.
    protein_db : dict of str, str
        A dictionary that maps protein accessions to their sequences.

    Returns
    -------
    tuple of list of str, list of int
        List of protein accessions, and list of peptide positions.

    Raises
    ------
    RuntimeError
        If the peptide could not be matched to any protein.

    Notes
    -----
    This function should not be called directly, it is called from ``reannotate_positions()``.

    Warnings
    --------
    Contrary to most functions in pyXLMS, this function uses 0-based indexing.
    """
    proteins = list()
    positions = list()
    for id, base_seq in protein_db.items():
        seqs = __generate_all_sequences(base_seq)
        for seq in seqs:
            if peptide in seq:
                for match in re.finditer(peptide, seq):
                    proteins.append(id)
                    positions.append(match.start())
    if len(proteins) == 0:
        raise RuntimeError(f"No match found for peptide {peptide}!")
    return (proteins, positions)


def fasta_title_to_accession(title: str) -> str:
    r"""Parses the protein accession from a UniProt-like title.

    Parameters
    ----------
    title : str
        Fasta title/header.

    Returns
    -------
    str
        The protein accession parsed from the title. If parsing was unsuccessful
        the full title is returned.

    Examples
    --------
    >>> from pyXLMS.transform import fasta_title_to_accession
    >>> title = "sp|A0A087X1C5|CP2D7_HUMAN Putative cytochrome P450 2D7 OS=Homo sapiens OX=9606 GN=CYP2D7 PE=5 SV=1"
    >>> fasta_title_to_accession(title)
    'A0A087X1C5'

    >>> from pyXLMS.transform import fasta_title_to_accession
    >>> title = "Cas9"
    >>> fasta_title_to_accession(title)
    'Cas9'
    """
    if "|" in title:
        return title.split("|")[1].strip()
    return title.strip()


def reannotate_positions(
    data: List[Dict[str, Any]] | Dict[str, Any],
    fasta: str | BinaryIO,
    title_to_accession: Optional[Callable[[str], str]] = None,
) -> List[Dict[str, Any]] | Dict[str, Any]:
    r"""Reannotates protein crosslink positions for a given fasta file.

    Reannotates the crosslink and peptide positions of the given cross-linked peptide pair and
    the specified fasta file. Takes a list of crosslink-spectrum-matches or crosslinks, or a
    parser_result as input.

    Parameters
    ----------
    data : list of dict of str, any, or dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to annotate, or a parser_result.
    fasta : str, or file stream
        The name/path of the fasta file containing protein sequences or a file-like object/stream.
    title_to_accession : callable, or None, default = None
        A function that parses the protein accession from the fasta title/header. If None (default)
        the function ``fasta_title_to_accession`` is used.

    Returns
    -------
    list of dict of str, any, or dict of str, any
        If a list of crosslink-spectrum-matches or crosslinks was provided, a list of annotated
        crosslink-spectrum-matches or crosslinks is returned. If a parser_result was provided,
        an annotated parser_result will be returned.

    Raises
    ------
    TypeError
        If a wrong data type is provided.

    Examples
    --------
    >>> from pyXLMS.data import create_crosslink_min
    >>> from pyXLMS.transform import reannotate_positions
    >>> xls = [create_crosslink_min("ADANLDK", 7, "GNTDRHSIK", 9)]
    >>> xls = reannotate_positions(xls, "data/_fasta/Cas9_plus10.fasta")
    >>> xls[0]["alpha_proteins"]
    ["Cas9"]
    >>> xls[0]["alpha_proteins_crosslink_positions"]
    [1293]
    >>> xls[0]["beta_proteins"]
    ["Cas9"]
    >>> xls[0]["beta_proteins_crosslink_positions"]
    [48]
    """
    if title_to_accession is not None:
        _ok = check_input(title_to_accession, "title_to_accession", Callable)
    else:
        title_to_accession = fasta_title_to_accession
    if isinstance(data, list):
        _ok = check_input(data, "data", list, dict)
        if len(data) == 0:
            return data
        if "data_type" not in data[0]:
            raise TypeError(
                "Can't annotate positions for input data. Input data has to be a list of crosslink-spectrum-matches or crosslinks "
                "or a 'parser_result'!"
            )
        _ok = assert_data_type_same(data)
        protein_db = dict()
        reannoted = list()
        # read fasta file
        i = 0
        if isinstance(fasta, str):
            with open(fasta, "r", encoding="utf-8") as f:
                for i, item in enumerate(SimpleFastaParser(f)):
                    protein_db[title_to_accession(item[0])] = item[1]
            if len(protein_db) != i + 1:
                warnings.warn(
                    RuntimeWarning(
                        f"Possible duplicates found in fasta file! Read {i + 1} sequences but only stored {len(protein_db)}."
                    )
                )
        else:
            for i, item in enumerate(SimpleFastaParser(fasta)):
                protein_db[title_to_accession(item[0])] = item[1]
            if len(protein_db) != i + 1:
                warnings.warn(
                    RuntimeWarning(
                        f"Possible duplicates found in fasta file! Read {i + 1} sequences but only stored {len(protein_db)}."
                    )
                )
        # annotate crosslinks
        if data[0]["data_type"] == "crosslink":
            for xl in tqdm(data, total=len(data), desc="Annotating crosslinks..."):
                proteins_a, pep_position0_proteins_a = __get_proteins_and_positions(
                    xl["alpha_peptide"], protein_db
                )
                proteins_b, pep_position0_proteins_b = __get_proteins_and_positions(
                    xl["beta_peptide"], protein_db
                )
                reannoted.append(
                    create_crosslink(
                        peptide_a=xl["alpha_peptide"],
                        xl_position_peptide_a=xl["alpha_peptide_crosslink_position"],
                        proteins_a=proteins_a,
                        xl_position_proteins_a=[
                            pos + xl["alpha_peptide_crosslink_position"]
                            for pos in pep_position0_proteins_a
                        ],
                        decoy_a=xl["alpha_decoy"],
                        peptide_b=xl["beta_peptide"],
                        xl_position_peptide_b=xl["beta_peptide_crosslink_position"],
                        proteins_b=proteins_b,
                        xl_position_proteins_b=[
                            pos + xl["beta_peptide_crosslink_position"]
                            for pos in pep_position0_proteins_b
                        ],
                        decoy_b=xl["beta_decoy"],
                        score=xl["score"],
                        additional_information=xl["additional_information"],
                    )
                )
        # annotate csms
        elif data[0]["data_type"] == "crosslink-spectrum-match":
            for csm in tqdm(
                data, total=len(data), desc="Annotation crosslink-spectrum-matches..."
            ):
                proteins_a, pep_position0_proteins_a = __get_proteins_and_positions(
                    csm["alpha_peptide"], protein_db
                )
                proteins_b, pep_position0_proteins_b = __get_proteins_and_positions(
                    csm["beta_peptide"], protein_db
                )
                reannoted.append(
                    create_csm(
                        peptide_a=csm["alpha_peptide"],
                        modifications_a=csm["alpha_modifications"],
                        xl_position_peptide_a=csm["alpha_peptide_crosslink_position"],
                        proteins_a=proteins_a,
                        xl_position_proteins_a=[
                            pos + csm["alpha_peptide_crosslink_position"]
                            for pos in pep_position0_proteins_a
                        ],
                        pep_position_proteins_a=[
                            pos + 1 for pos in pep_position0_proteins_a
                        ],
                        score_a=csm["alpha_score"],
                        decoy_a=csm["alpha_decoy"],
                        peptide_b=csm["beta_peptide"],
                        modifications_b=csm["beta_modifications"],
                        xl_position_peptide_b=csm["beta_peptide_crosslink_position"],
                        proteins_b=proteins_b,
                        xl_position_proteins_b=[
                            pos + csm["beta_peptide_crosslink_position"]
                            for pos in pep_position0_proteins_b
                        ],
                        pep_position_proteins_b=[
                            pos + 1 for pos in pep_position0_proteins_b
                        ],
                        score_b=csm["beta_score"],
                        decoy_b=csm["beta_decoy"],
                        score=csm["score"],
                        spectrum_file=csm["spectrum_file"],
                        scan_nr=csm["scan_nr"],
                        charge=csm["charge"],
                        rt=csm["retention_time"],
                        im_cv=csm["ion_mobility"],
                        additional_information=csm["additional_information"],
                    )
                )
        else:
            raise TypeError(
                f"Can't annotate positions for data type {data[0]['data_type']}. Valid data types are:\n"
                "'crosslink-spectrum-match', 'crosslink', and 'parser_result'."
            )
        return reannoted
    _ok = check_input(data, "data", dict)
    if "data_type" not in data or data["data_type"] != "parser_result":
        raise TypeError(
            "Can't annotate positions for dict. Dict has to be a valid 'parser_result'!"
        )
    new_csms = (
        reannotate_positions(
            data["crosslink-spectrum-matches"], fasta, title_to_accession
        )
        if data["crosslink-spectrum-matches"] is not None
        else None
    )
    new_xls = (
        reannotate_positions(data["crosslinks"], fasta, title_to_accession)
        if data["crosslinks"] is not None
        else None
    )
    if new_csms is not None:
        if not isinstance(new_csms, list):
            raise RuntimeError(
                "Something went wrong while reannotating positions.\n"
                f"Expected data type: list. Got: {type(new_csms)}."
            )
    if new_xls is not None:
        if not isinstance(new_xls, list):
            raise RuntimeError(
                "Something went wrong while reannotating positions.\n"
                f"Expected data type: list. Got: {type(new_xls)}."
            )
    return create_parser_result(
        search_engine=data["search_engine"], csms=new_csms, crosslinks=new_xls
    )



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.exporter import to_xlinkdb
    from pyXLMS.parser import read

    pr = read(
        "data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv",
        engine="xiSearch/xiFDR",
        crosslinker="DSS",
    )
    crosslinks = pr["crosslinks"]
    df = to_xlinkdb(crosslinks, filename="crosslinksForXLinkDB")
    assert df.shape[0] == 227
    assert df.shape[1] == 7


def test2():
    from pyXLMS.exporter import to_xlinkdb
    from pyXLMS.parser import read

    pr = read(
        "data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv",
        engine="xiSearch/xiFDR",
        crosslinker="DSS",
    )
    crosslinks = pr["crosslinks"]
    df = to_xlinkdb(crosslinks, filename=None)
    assert df.shape[0] == 227
    assert df.shape[1] == 7


def test3():
    from pyXLMS.exporter import to_xlinkdb
    from pyXLMS.parser import read

    pr = read(
        "data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv",
        engine="xiSearch/xiFDR",
        crosslinker="DSS",
    )
    crosslinks = pr["crosslinks"]

    with pytest.raises(
        ValueError,
        match="Parameter filename must only contain alpha-numeric characters and no file extension!",
    ):
        _df = to_xlinkdb(crosslinks, filename="crosslinksForXLinkDB.tsv")


def test4():
    from pyXLMS.exporter import to_xlinkdb

    with pytest.raises(
        ValueError,
        match="Provided crosslinks contain no elements!",
    ):
        _df = to_xlinkdb([], filename="crosslinksForXLinkDB")


def test5():
    from pyXLMS.exporter import to_xlinkdb
    from pyXLMS.data import create_crosslink_min

    xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    crosslinks = [xl1, xl2]

    with pytest.raises(
        RuntimeError,
        match="Can't export to XLinkDB because not all necessary information is available!",
    ):
        _df = to_xlinkdb(crosslinks, filename=None)



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd

from ..data import check_input
from ..transform.util import assert_data_type_same
from .util import __get_filename
from .to_msannika import get_msannika_crosslink_sequence

from typing import Optional
from typing import Dict
from typing import Any
from typing import List


def __xls_to_xmas(
    xls: List[Dict[str, Any]],
    filename: Optional[str],
) -> pd.DataFrame:
    r"""Exports crosslinks to XMAS format.

    Parameters
    ----------
    xls : list of dict of str, any
        A list of crosslinks.
    filename : str, or None
        If not None, the data will be written to a file with the specified filename.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame in XMAS format.

    Notes
    -----
    This function should not be called directly, it is called from ``to_xmas()``.
    """
    sequence_a = list()
    sequence_b = list()
    for xl in xls:
        sequence_a.append(
            get_msannika_crosslink_sequence(
                xl["alpha_peptide"], xl["alpha_peptide_crosslink_position"]
            )
        )
        sequence_b.append(
            get_msannika_crosslink_sequence(
                xl["beta_peptide"], xl["beta_peptide_crosslink_position"]
            )
        )
    xmas_df = pd.DataFrame(
        {
            "Sequence A": sequence_a,
            "Sequence B": sequence_b,
        }
    )
    if filename is not None:
        xmas_df.to_excel(
            __get_filename(filename, "xlsx"), engine="openpyxl", index=False
        )
    return xmas_df


def to_xmas(
    crosslinks: List[Dict[str, Any]],
    filename: Optional[str],
) -> pd.DataFrame:
    r"""Exports a list of crosslinks to XMAS format.

    Exports a list of crosslinks to XMAS format for visualization in ChimeraX. The tool XMAS
    is available from
    `github.com/ScheltemaLab/ChimeraX_XMAS_bundle <https://github.com/ScheltemaLab/ChimeraX_XMAS_bundle>`_.

    Parameters
    ----------
    crosslinks : list of dict of str, any
        A list of crosslinks.
    filename : str, or None
        If not None, the exported data will be written to a file with the specified filename.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing crosslinks in XMAS format.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If 'crosslinks' parameter contains elements of mixed data type.
    ValueError
        If the provided 'crosslinks' parameter contains no elements.

    Examples
    --------
    >>> from pyXLMS.exporter import to_xmas
    >>> from pyXLMS.data import create_crosslink_min
    >>> xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    >>> xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    >>> crosslinks = [xl1, xl2]
    >>> to_xmas(crosslinks, filename="crosslinks_xmas.xlsx")
       Sequence A  Sequence B
    0  [K]PEPTIDE  P[K]EPTIDE
    1  PE[K]PTIDE  PEP[K]TIDE

    >>> from pyXLMS.exporter import to_xmas
    >>> from pyXLMS.data import create_crosslink_min
    >>> xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    >>> xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    >>> crosslinks = [xl1, xl2]
    >>> to_xmas(crosslinks, filename=None)
       Sequence A  Sequence B
    0  [K]PEPTIDE  P[K]EPTIDE
    1  PE[K]PTIDE  PEP[K]TIDE
    """
    _ok = check_input(crosslinks, "crosslinks", list, dict)
    _ok = check_input(filename, "filename", str) if filename is not None else True
    if len(crosslinks) == 0:
        raise ValueError("Provided crosslinks contain no elements!")
    if "data_type" not in crosslinks[0] or crosslinks[0]["data_type"] != "crosslink":
        raise TypeError(
            "Unsupported data type for input crosslinks! Parameter crosslinks has to be a list of crosslinks!"
        )
    if not assert_data_type_same(crosslinks):
        raise TypeError("Not all elements in data have the same data type!")
    return __xls_to_xmas(crosslinks, filename)



#!/usr/bin/env python3

# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

AMINO_ACIDS = {
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
    # non-standard
    "U",
    "O",
    # placeholders
    "X",
    "B",
    "Z",
    "J",
}
r"""List of valid amino acids.

List of one-letter codes for all valid amino acids.

Examples
--------
>>> from pyXLMS.constants import AMINO_ACIDS
>>> "A" in AMINO_ACIDS
True
>>> "B" in AMINO_ACIDS
True
"""

AMINO_ACIDS_REPLACEMENTS = {
    "X": {
        "A",
        "R",
        "N",
        "D",
        "C",
        "E",
        "Q",
        "G",
        "H",
        "I",
        "L",
        "K",
        "M",
        "F",
        "P",
        "S",
        "T",
        "W",
        "Y",
        "V",
        # non-standard
        "U",
        "O",
    },
    "B": {"D", "N"},
    "Z": {"E", "Q"},
    "J": {"I", "L"},
}
r"""Mapping of placeholder amino acid codes to their respective possible amino acids.

Dictionary mapping placeholder one-letter amino acid codes to their respective possible amino acids

Examples
--------
>>> from pyXLMS.constants import AMINO_ACIDS_REPLACEMENTS
>>> AMINO_ACIDS_REPLACEMENTS["B"]
{'D', 'N'}
>>> AMINO_ACIDS_REPLACEMENTS["Z"]
{'E', 'Q'}
"""

AMINO_ACIDS_3TO1 = {
    "GLY": "G",
    "PRO": "P",
    "ALA": "A",
    "VAL": "V",
    "LEU": "L",
    "ILE": "I",
    "MET": "M",
    "CYS": "C",
    "PHE": "F",
    "TYR": "Y",
    "TRP": "W",
    "HIS": "H",
    "LYS": "K",
    "ARG": "R",
    "GLN": "Q",
    "ASN": "N",
    "GLU": "E",
    "ASP": "D",
    "SER": "S",
    "THR": "T",
    # non-standard
    "SEC": "U",
    "PYL": "O",
    # placeholders
    "XAA": "X",
    "ASX": "B",
    "GLX": "Z",
    "XLE": "J",
}
r"""Mapping of amino acid 3-letter codes to their 1-letter codes.

Mapping of all amino acid 3-letter codes to their corresponding 1-letter codes.

Examples
--------
>>> from pyXLMS.constants import AMINO_ACIDS_3TO1
>>> AMINO_ACIDS_3TO1["GLY"]
'G'
"""

AMINO_ACIDS_1TO3 = {v: k for k, v in AMINO_ACIDS_3TO1.items()}
r"""Mapping of amino acid 1-letter codes to their 3-letter codes.

Mapping of all amino acid 1-letter codes to their corresponding 3-letter codes.

Examples
--------
>>> from pyXLMS.constants import AMINO_ACIDS_1TO3
>>> AMINO_ACIDS_1TO3["G"]
'GLY'
"""

CROSSLINKERS = {
    "BS3": 138.06808,
    "DSS": 138.06808,
    "DSSO": 158.00376,
    "DSBU": 196.08479231,
    "ADH": 138.09054635,
    "DSBSO": 308.03883,
    "PhoX": 209.97181,
    "DSG": 96.0211293726,
}
r"""Dictionary of crosslinkers.

Dictionary of pre-defined crosslinkers that maps crosslinker names to crosslinker delta masses.
Currently contains `"BS3"`, `"DSS"`, `"DSSO"`, `"DSBU"`, `"ADH"`, `"DSBSO"`, `"PhoX"`, `"DSG"`.

Examples
--------
>>> from pyXLMS.constants import CROSSLINKERS
>>> CROSSLINKERS["BS3"]
138.06808
"""

MODIFICATIONS = {
    "Carbamidomethyl": 57.021464,
    "Oxidation": 15.994915,
    "Phospho": 79.966331,
    "Acetyl": 42.010565,
}
r"""Dictionary of post-translational-modifications.

Dictionary of pre-defined post-translational-modifications that maps modification names to modification delta masses.
Currently contains `"Carbamidomethyl"`, `"Oxidation"`, `"Phospho"`, `"Acetyl"` and all crosslinkers.

Examples
--------
>>> from pyXLMS.constants import MODIFICATIONS
>>> MODIFICATIONS["Carbamidomethyl"]
57.021464
>>> MODIFICATIONS["BS3"]
138.06808
"""

MODIFICATIONS.update(CROSSLINKERS)

XI_MODIFICATION_MAPPING = {
    "->": ("Substitution", float("nan")),
    "cm": ("Carbamidomethyl", 57.021464),
    "ox": ("Oxidation", 15.994915),
    "bs3oh": ("BS3 Hydrolized", 156.0786347),
    "bs3nh2": ("BS3 Amidated", 155.094619105),
    "bs3loop": ("BS3 Looplink", 138.06808),
    "bs3_hyd": ("BS3 Hydrolized", 156.0786347),
    "bs3_ami": ("BS3 Amidated", 155.094619105),
    "bs3_tris": ("BS3 Tris", 259.141973),
    "dssoloop": ("DSSO Looplink", 158.00376),
    "dsso_loop": ("DSSO Looplink", 158.00376),
    "dsso_hyd": ("DSSO Hydrolized", 176.0143295),
    "dsso_ami": ("DSSO Amidated", 175.030313905),
    "dsso_tris": ("DSSO Tris", 279.077658),
    "dsbuloop": ("DSBU Looplink", 196.08479231),
    "dsbu_loop": ("DSBU Looplink", 196.08479231),
    "dsbu_hyd": ("DSBU Hydrolized", 214.095357),
    "dsbu_ami": ("DSBU Amidated", 213.111341),
    "dsbu_tris": ("DSBU Tris", 317.158685),
}
r"""Dictionary that maps sequence elements from xiSearch and xiFDR to their corresponding post-translational-modifications.

Dictionary that maps sequence elements (e.g. `"cm"`) from xiSearch and xiFDR to their corresponding
post-translational-modifications (e.g. `("Carbamidomethyl", 57.021464)`).

Examples
--------
>>> from pyXLMS.constants import XI_MODIFICATION_MAPPING
>>> XI_MODIFICATION_MAPPING["cm"]
('Carbamidomethyl', 57.021464)
>>> XI_MODIFICATION_MAPPING["ox"]
('Oxidation', 15.994915)
"""

SCOUT_MODIFICATION_MAPPING = {
    "+57.021460": ("Carbamidomethyl", 57.021464),
    "+15.994900": ("Oxidation", 15.994915),
    "Oxidation of Methionine": ("Oxidation", 15.994915),
    "Carbamidomethyl": ("Carbamidomethyl", 57.021464),
}
r"""Dictionary that maps sequence elements and modifications from Scout to their corresponding post-translational-modifications.

Dictionary that maps sequence elements (e.g. `"+57.021460"`) and modifications (e.g. `"Carbamidomethyl"`) from Scout to their
corresponding post-translational-modifications (e.g. `("Carbamidomethyl", 57.021464)`).

Examples
--------
>>> from pyXLMS.constants import SCOUT_MODIFICATION_MAPPING
>>> SCOUT_MODIFICATION_MAPPING["+57.021460"]
('Carbamidomethyl', 57.021464)
>>> SCOUT_MODIFICATION_MAPPING["Carbamidomethyl"]
('Carbamidomethyl', 57.021464)
>>> SCOUT_MODIFICATION_MAPPING["Oxidation of Methionine"]
('Oxidation', 15.994915)
"""

SCOUT_CROSSLINKER_MAPPING = {k: (k, v) for k, v in CROSSLINKERS.items()}
SCOUT_MODIFICATION_MAPPING.update(SCOUT_CROSSLINKER_MAPPING)

MEROX_MODIFICATION_MAPPING = {
    "B": {"Amino Acid": "C", "Modification": ("Carbamidomethyl", 57.021464)},
    "m": {"Amino Acid": "M", "Modification": ("Oxidation", 15.994915)},
}
r"""Dictionary that maps MeroX modification symbols to their corresponding amino acids and post-translational-modifications.

Dictionary that maps MeroX modification symbols (e.g. `"B"`) to their corresponding amino acids and post-translational-modifications
(e.g. ``{"Amino Acid": "C", "Modification": ("Carbamidomethyl", 57.021464)}``).

Examples
--------
>>> from pyXLMS.constants import MEROX_MODIFICATION_MAPPING
>>> MEROX_MODIFICATION_MAPPING["B"]
{'Amino Acid': 'C', 'Modification': ('Carbamidomethyl', 57.021464)}

>>> from pyXLMS.constants import MEROX_MODIFICATION_MAPPING
>>> MEROX_MODIFICATION_MAPPING["m"]
{'Amino Acid': 'M', 'Modification': ('Oxidation', 15.994915)}

>>> from pyXLMS.constants import MEROX_MODIFICATION_MAPPING
>>> MEROX_MODIFICATION_MAPPING
{'B': {'Amino Acid': 'C', 'Modification': ('Carbamidomethyl', 57.021464)}, 'm': {'Amino Acid': 'M', 'Modification': ('Oxidation', 15.994915)}}
"""



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


@pytest.fixture(autouse=True)
def cleanup_figures():
    from matplotlib import use
    import matplotlib.pyplot as plt

    use("agg")

    yield

    plt.close(fig="all")


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test1():
    from pyXLMS.plotting import venn

    fig, ax = venn(
        {"A", "B", "C"},
        {"B", "C", "D", "E", "F"},
        labels=["A", "F"],
        colors=["orange", "blue"],
    )
    assert fig is not None
    assert ax is not None


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test2():
    from pyXLMS.plotting import venn

    fig, ax = venn({"A", "B", "C"}, {"B", "C", "D", "E", "F"}, {"F", "G"})
    assert fig is not None
    assert ax is not None


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test3():
    from pyXLMS.plotting import venn

    with pytest.warns(
        RuntimeWarning,
        match=r"More than two labels supplied for two sets\. Using first two\.\.\.",
    ):
        _fig, _ax = venn(
            {"A", "B", "C"}, {"B", "C", "D", "E", "F"}, colors=["orange", "blue"]
        )


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test4():
    from pyXLMS.plotting import venn

    with pytest.warns(
        RuntimeWarning,
        match=r"More than two colors supplied for two sets\. Using first two\.\.\.",
    ):
        _fig, _ax = venn({"A", "B", "C"}, {"B", "C", "D", "E", "F"}, labels=["A", "F"])


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test5():
    from pyXLMS.plotting import venn

    with pytest.raises(
        IndexError,
        match=r"At least two labels have to be given if two sets are supplied!",
    ):
        _fig, _ax = venn(
            {"A", "B", "C"},
            {"B", "C", "D", "E", "F"},
            labels=["A"],
            colors=["orange", "blue"],
        )


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test6():
    from pyXLMS.plotting import venn

    with pytest.raises(
        IndexError,
        match=r"At least two colors have to be given if two sets are supplied!",
    ):
        _fig, _ax = venn(
            {"A", "B", "C"},
            {"B", "C", "D", "E", "F"},
            labels=["A", "F"],
            colors=["orange"],
        )


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test7():
    from pyXLMS.plotting import venn

    with pytest.warns(
        RuntimeWarning,
        match=r"More than three colors supplied for three sets\. Using first three\.\.\.",
    ):
        _fig, _ax = venn(
            {"A", "B", "C"},
            {"B", "C", "D", "E", "F"},
            {"F", "G"},
            colors=["orange", "blue", "green", "red"],
        )


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test8():
    from pyXLMS.plotting import venn

    with pytest.warns(
        RuntimeWarning,
        match=r"More than three labels supplied for three sets\. Using first three\.\.\.",
    ):
        _fig, _ax = venn(
            {"A", "B", "C"},
            {"B", "C", "D", "E", "F"},
            {"F", "G"},
            labels=["A", "B", "C", "D"],
        )


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test9():
    from pyXLMS.plotting import venn

    with pytest.raises(
        IndexError,
        match=r"At least three labels have to be given if three sets are supplied!",
    ):
        _fig, _ax = venn(
            {"A", "B", "C"}, {"B", "C", "D", "E", "F"}, {"F", "G"}, labels=["A", "B"]
        )


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test10():
    from pyXLMS.plotting import venn

    with pytest.raises(
        IndexError,
        match=r"At least three colors have to be given if three sets are supplied!",
    ):
        _fig, _ax = venn(
            {"A", "B", "C"},
            {"B", "C", "D", "E", "F"},
            {"F", "G"},
            colors=["orange", "blue"],
        )



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.transform import to_proforma

    psm = {"peptide": "PEPKTIDE"}

    with pytest.raises(
        TypeError,
        match=r"Unsupported data type for input data! Parameter data has to be a \(list of\) crosslink or crosslink-spectrum-match!",
    ):
        _r = to_proforma(psm)


def test2():
    from pyXLMS.data import create_crosslink_min
    from pyXLMS.transform import to_proforma

    xl = create_crosslink_min("PEPKTIDE", 4, "KPEPTIDE", 1)
    assert to_proforma(xl) == "KPEPTIDE//PEPKTIDE"


def test3():
    from pyXLMS.data import create_crosslink_min
    from pyXLMS.transform import to_proforma

    xl = create_crosslink_min("PEPKTIDE", 4, "KPEPTIDE", 1)
    assert (
        to_proforma(xl, crosslinker="Xlink:DSSO")
        == "K[Xlink:DSSO]PEPTIDE//PEPK[Xlink:DSSO]TIDE"
    )


def test4():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min("PEPKTIDE", 4, "KPEPTIDE", 1, "RUN_1", 1)
    assert to_proforma(csm) == "KPEPTIDE//PEPKTIDE"


def test5():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min("PEPKTIDE", 4, "KPEPTIDE", 1, "RUN_1", 1)
    assert (
        to_proforma(csm, crosslinker="Xlink:DSSO")
        == "K[Xlink:DSSO]PEPTIDE//PEPK[Xlink:DSSO]TIDE"
    )


def test6():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min(
        "PEPKTIDE",
        4,
        "KPMEPTIDE",
        1,
        "RUN_1",
        1,
        modifications_b={3: ("Oxidation", 15.994915)},
    )
    assert (
        to_proforma(csm, crosslinker="Xlink:DSSO")
        == "K[Xlink:DSSO]PM[+15.994915]EPTIDE//PEPK[Xlink:DSSO]TIDE"
    )


def test7():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min(
        "PEPKTIDE",
        4,
        "KPMEPTIDE",
        1,
        "RUN_1",
        1,
        modifications_b={3: ("Oxidation", 15.994915)},
        charge=3,
    )
    assert (
        to_proforma(csm, crosslinker="Xlink:DSSO")
        == "K[Xlink:DSSO]PM[+15.994915]EPTIDE//PEPK[Xlink:DSSO]TIDE/3"
    )


def test8():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min(
        "PEPKTIDE",
        4,
        "KPMEPTIDE",
        1,
        "RUN_1",
        1,
        modifications_a={4: ("DSSO", 158.00376)},
        modifications_b={1: ("DSSO", 158.00376), 3: ("Oxidation", 15.994915)},
        charge=3,
    )
    assert (
        to_proforma(csm) == "K[+158.00376]PM[+15.994915]EPTIDE//PEPK[+158.00376]TIDE/3"
    )


def test9():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min(
        "PEPKTIDE",
        4,
        "KPMEPTIDE",
        1,
        "RUN_1",
        1,
        modifications_a={4: ("DSSO", 158.00376)},
        modifications_b={1: ("DSSO", 158.00376), 3: ("Oxidation", 15.994915)},
        charge=3,
    )
    assert (
        to_proforma(csm, crosslinker="Xlink:DSSO")
        == "K[+158.00376]PM[+15.994915]EPTIDE//PEPK[+158.00376]TIDE/3"
    )


def test10():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min(
        "PEPKTIDE",
        4,
        "MKPMEPTIDE",
        2,
        "RUN_1",
        1,
        modifications_a={4: ("DSSO", 158.00376)},
        modifications_b={
            2: ("DSSO", 158.00376),
            4: ("Oxidation", 15.994915),
            1: ("Oxidation", 15.994915),
        },
        charge=3,
    )
    assert (
        to_proforma(csm)
        == "M[+15.994915]K[+158.00376]PM[+15.994915]EPTIDE//PEPK[+158.00376]TIDE/3"
    )


def test11():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm_1 = create_csm_min(
        "PEPKTIDE",
        4,
        "KPMEPTIDE",
        1,
        "RUN_1",
        1,
        modifications_a={4: ("DSSO", 158.00376)},
        modifications_b={1: ("DSSO", 158.00376), 3: ("Oxidation", 15.994915)},
        charge=3,
    )
    csm_2 = create_csm_min(
        "PEPKTIDE",
        4,
        "MKPMEPTIDE",
        2,
        "RUN_1",
        2,
        modifications_a={4: ("DSSO", 158.00376)},
        modifications_b={
            2: ("DSSO", 158.00376),
            4: ("Oxidation", 15.994915),
            1: ("Oxidation", 15.994915),
        },
        charge=3,
    )
    proforma = to_proforma([csm_1, csm_2])
    assert proforma[0] == "K[+158.00376]PM[+15.994915]EPTIDE//PEPK[+158.00376]TIDE/3"
    assert (
        proforma[1]
        == "M[+15.994915]K[+158.00376]PM[+15.994915]EPTIDE//PEPK[+158.00376]TIDE/3"
    )


def test12():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma
    from pyXLMS.transform import modifications_to_str as mts

    csm = create_csm_min(
        "PEPKTIDE",
        4,
        "KPMEPTIDE",
        1,
        "RUN_1",
        1,
        modifications_b={3: ("Oxidation", 15.994915)},
        charge=3,
    )
    assert (
        to_proforma(csm, crosslinker="Xlink:DSSO")
        == "K[Xlink:DSSO]PM[+15.994915]EPTIDE//PEPK[Xlink:DSSO]TIDE/3"
    )
    assert mts(csm["alpha_modifications"]) == "(3:[Oxidation|15.994915])"


def test13():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min(
        "PEPKTIDE",
        4,
        "MKPMEPTIDE",
        1,
        "RUN_1",
        1,
        modifications_a={4: ("DSSO", 158.00376)},
        modifications_b={
            0: ("DSSO", 158.00376),
            4: ("Oxidation", 15.994915),
            1: ("Oxidation", 15.994915),
        },
        charge=3,
    )
    assert (
        to_proforma(csm)
        == "[+158.00376]-M[+15.994915]KPM[+15.994915]EPTIDE//PEPK[+158.00376]TIDE/3"
    )


def test14():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min(
        "PEPKTIDE",
        4,
        "MKPMEPTIDE",
        10,
        "RUN_1",
        1,
        modifications_a={4: ("DSSO", 158.00376)},
        modifications_b={
            11: ("DSSO", 158.00376),
            4: ("Oxidation", 15.994915),
            1: ("Oxidation", 15.994915),
        },
        charge=3,
    )
    assert (
        to_proforma(csm)
        == "M[+15.994915]KPM[+15.994915]EPTIDE-[+158.00376]//PEPK[+158.00376]TIDE/3"
    )


def test15():
    from pyXLMS.data import create_csm_min
    from pyXLMS.transform import to_proforma

    csm = create_csm_min(
        "PEPKTIDE",
        4,
        "MKPMEPTIDE",
        10,
        "RUN_1",
        1,
        modifications_a={4: ("DSSO", 158.00376)},
        modifications_b={
            11: ("DSSO", 158.00376),
            4: ("Oxidation", 15.994915),
            1: ("Oxidation", float("nan")),
        },
        charge=3,
    )
    assert (
        to_proforma(csm)
        == "MKPM[+15.994915]EPTIDE-[+158.00376]//PEPK[+158.00376]TIDE/3"
    )



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd
from tqdm import tqdm

from ..data import check_input
from ..data import create_crosslink
from ..data import create_csm
from ..parser.parser_xldbse_custom import __get_value
from ..parser.parser_xldbse_custom import pyxlms_modification_str_parser
from ..parser.util import format_sequence
from ..parser.util import get_bool_from_value
from ..parser.util import __serialize_pandas_series

from typing import Optional
from typing import Dict
from typing import Any
from typing import Tuple
from typing import List
from typing import Callable


def from_dataframe(
    df: pd.DataFrame,
    column_mapping: Optional[Dict[str, str]] = None,
    parse_modifications: bool = True,
    modification_parser: Optional[Callable[[str], Dict[int, Tuple[str, float]]]] = None,
    decoy_prefix: str = "REV_",
) -> List[Dict[str, Any]]:
    r"""Read a pandas DataFrame in custom or pyXLMS format.

    Reads a pandas DataFrame in custom or pyXLMS format and returns a list of crosslink-spectrum-matches or crosslinks.

    The minimum required columns for a crosslink-spectrum-matches pandas DataFrame are:

    - "Alpha Peptide": The unmodified amino acid sequence of the first peptide.
    - "Alpha Peptide Crosslink Position": The position of the crosslinker in the sequence of the first peptide (1-based).
    - "Beta Peptide": The unmodified amino acid sequence of the second peptide.
    - "Beta Peptide Crosslink Position": The position of the crosslinker in the sequence of the second peptide (1-based).
    - "Spectrum File": Name of the spectrum file the crosslink-spectrum-match was identified in.
    - "Scan Nr": The corresponding scan number of the crosslink-spectrum-match.

    The minimum required columns for crosslink pandas DataFrame are:

    - "Alpha Peptide": The unmodified amino acid sequence of the first peptide.
    - "Alpha Peptide Crosslink Position": The position of the crosslinker in the sequence of the first peptide (1-based).
    - "Beta Peptide": The unmodified amino acid sequence of the second peptide.
    - "Beta Peptide Crosslink Position": The position of the crosslinker in the sequence of the second peptide (1-based).

    A full specification of columns that can be parsed can be found in the
    `docs <https://github.com/hgb-bin-proteomics/pyXLMS/blob/master/docs/format.md>`_.

    Parameters
    ----------
    df : pandas.DataFrame
        The pandas.DataFrame containing crosslink-spectrum-matches or crosslinks.
    column_mapping : dict of str, str
        A dictionary that maps the ``df`` columns to the required pyXLMS column names.
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modification_parser' parameter.
    modification_parser : callable, or None
        A function that parses modification strings and returns the pyXLMS specific modifications object.
        If None, the function ``pyxlms_modification_str_parser()`` is used. If no modification columns are
        given this parameter is ignored.
    decoy_prefix : str, default = "REV\_"
        The prefix that indicates that a protein is from the decoy database.

    Returns
    -------
    list of dict of str, any
        If a crosslink-spectrum-matches DataFrame was given, a list of crosslink-spectrum-matches is returned.
        If a crosslinks DataFrame was given, a list of crosslinks is returned.

    Raises
    ------
    TypeError
        If one of the values could not be parsed.

    Examples
    --------
    >>> from pyXLMS import parser, transform
    >>> csms_from_pyxlms = parser.read_custom("data/pyxlms/csm.txt")
    >>> csms_df = transform.to_dataframe(csms_from_pyxlms["crosslink-spectrum-matches"])
    >>> csms_from_pyxlms = transform.from_dataframe(csms_df)

    >>> from pyXLMS import parser, transform
    >>> crosslinks_from_pyxlms = parser.read_custom("data/pyxlms/xl.txt")
    >>> crosslinks_df = transform.to_dataframe(crosslinks_from_pyxlms["crosslinks"])
    >>> crosslinks_from_pyxlms = transform.from_dataframe(crosslinks_df)
    """
    ## check input
    _ok = check_input(df, "df", pd.DataFrame)
    _ok = (
        check_input(column_mapping, "column_mapping", dict, str)
        if column_mapping is not None
        else True
    )
    _ok = check_input(parse_modifications, "parse_modifications", bool)
    _ok = (
        check_input(modification_parser, "modification_parser", Callable)
        if modification_parser is not None
        else True
    )
    _ok = check_input(decoy_prefix, "decoy_prefix", str)
    ## helper functions

    def get_is_decoy_value(
        row: pd.Series, decoy_prefix: str, alpha: bool
    ) -> bool | None:
        if alpha:
            if __get_value(row, "Alpha Decoy") is not None:
                return get_bool_from_value(__get_value(row, "Alpha Decoy"))
            if __get_value(row, "Alpha Proteins") is not None:
                return decoy_prefix in str(__get_value(row, "Alpha Proteins"))
            return None
        if __get_value(row, "Beta Decoy") is not None:
            return get_bool_from_value(__get_value(row, "Beta Decoy"))
        if __get_value(row, "Beta Proteins") is not None:
            return decoy_prefix in str(__get_value(row, "Beta Proteins"))
        return None

    def get_int(value: Any) -> int | None:
        if value is None:
            return None
        try:
            return int(value)
        except Exception as _e:
            pass
        raise TypeError(f"Could not parse int from value {value}!")
        return None

    def get_float(value: Any) -> float | None:
        if value is None:
            return None
        try:
            return float(value)
        except Exception as _e:
            pass
        raise TypeError(f"Could not parse float from value {value}!")
        return None

    ## set default parser
    if modification_parser is None:
        modification_parser = pyxlms_modification_str_parser

    ## data structures
    crosslinks = list()
    csms = list()

    ## detect input file type
    data = df.copy(deep=True)
    if column_mapping is not None:
        data.rename(columns=column_mapping, inplace=True)
    col_names = data.columns.values.tolist()
    is_crosslink_dataframe = "Scan Nr" not in col_names
    ## process data
    if is_crosslink_dataframe:
        for i, row in tqdm(
            data.iterrows(),
            total=data.shape[0],
            desc="Reading crosslinks...",
        ):
            # create crosslink
            crosslink = create_crosslink(
                peptide_a=format_sequence(str(row["Alpha Peptide"])),
                xl_position_peptide_a=int(row["Alpha Peptide Crosslink Position"]),
                proteins_a=[
                    protein.strip()
                    if protein.strip()[: len(decoy_prefix)] != decoy_prefix
                    else protein.strip()[len(decoy_prefix) :]
                    for protein in str(__get_value(row, "Alpha Proteins")).split(";")
                ]
                if __get_value(row, "Alpha Proteins") is not None
                else None,
                xl_position_proteins_a=[
                    int(position)
                    for position in str(
                        __get_value(row, "Alpha Proteins Crosslink Positions")
                    ).split(";")
                ]
                if __get_value(row, "Alpha Proteins Crosslink Positions") is not None
                else None,
                decoy_a=get_is_decoy_value(row, decoy_prefix, True),
                peptide_b=format_sequence(str(row["Beta Peptide"])),
                xl_position_peptide_b=int(row["Beta Peptide Crosslink Position"]),
                proteins_b=[
                    protein.strip()
                    if protein.strip()[: len(decoy_prefix)] != decoy_prefix
                    else protein.strip()[len(decoy_prefix) :]
                    for protein in str(__get_value(row, "Beta Proteins")).split(";")
                ]
                if __get_value(row, "Beta Proteins") is not None
                else None,
                xl_position_proteins_b=[
                    int(position)
                    for position in str(
                        __get_value(row, "Beta Proteins Crosslink Positions")
                    ).split(";")
                ]
                if __get_value(row, "Beta Proteins Crosslink Positions") is not None
                else None,
                decoy_b=get_is_decoy_value(row, decoy_prefix, False),
                score=get_float(__get_value(row, "Crosslink Score")),
                additional_information={"source": __serialize_pandas_series(row)},
            )
            crosslinks.append(crosslink)
        return crosslinks
    for i, row in tqdm(
        data.iterrows(),
        total=data.shape[0],
        desc="Reading CSMs...",
    ):
        # create csm
        csm = create_csm(
            peptide_a=format_sequence(str(row["Alpha Peptide"])),
            modifications_a=modification_parser(
                str(__get_value(row, "Alpha Peptide Modifications"))
            )
            if parse_modifications
            and __get_value(row, "Alpha Peptide Modifications") is not None
            else None,
            xl_position_peptide_a=int(row["Alpha Peptide Crosslink Position"]),
            proteins_a=[
                protein.strip()
                if protein.strip()[: len(decoy_prefix)] != decoy_prefix
                else protein.strip()[len(decoy_prefix) :]
                for protein in str(__get_value(row, "Alpha Proteins")).split(";")
            ]
            if __get_value(row, "Alpha Proteins") is not None
            else None,
            xl_position_proteins_a=[
                int(position)
                for position in str(
                    __get_value(row, "Alpha Proteins Crosslink Positions")
                ).split(";")
            ]
            if __get_value(row, "Alpha Proteins Crosslink Positions") is not None
            else None,
            pep_position_proteins_a=[
                int(position)
                for position in str(
                    __get_value(row, "Alpha Proteins Peptide Positions")
                ).split(";")
            ]
            if __get_value(row, "Alpha Proteins Peptide Positions") is not None
            else None,
            score_a=get_float(__get_value(row, "Alpha Score")),
            decoy_a=get_is_decoy_value(row, decoy_prefix, True),
            peptide_b=format_sequence(str(row["Beta Peptide"])),
            modifications_b=modification_parser(
                str(__get_value(row, "Beta Peptide Modifications"))
            )
            if parse_modifications
            and __get_value(row, "Beta Peptide Modifications") is not None
            else None,
            xl_position_peptide_b=int(row["Beta Peptide Crosslink Position"]),
            proteins_b=[
                protein.strip()
                if protein.strip()[: len(decoy_prefix)] != decoy_prefix
                else protein.strip()[len(decoy_prefix) :]
                for protein in str(__get_value(row, "Beta Proteins")).split(";")
            ]
            if __get_value(row, "Beta Proteins") is not None
            else None,
            xl_position_proteins_b=[
                int(position)
                for position in str(
                    __get_value(row, "Beta Proteins Crosslink Positions")
                ).split(";")
            ]
            if __get_value(row, "Beta Proteins Crosslink Positions") is not None
            else None,
            pep_position_proteins_b=[
                int(position)
                for position in str(
                    __get_value(row, "Beta Proteins Peptide Positions")
                ).split(";")
            ]
            if __get_value(row, "Beta Proteins Peptide Positions") is not None
            else None,
            score_b=get_float(__get_value(row, "Beta Score")),
            decoy_b=get_is_decoy_value(row, decoy_prefix, False),
            score=get_float(__get_value(row, "CSM Score")),
            spectrum_file=str(row["Spectrum File"]).strip(),
            scan_nr=int(row["Scan Nr"]),
            charge=get_int(__get_value(row, "Precursor Charge")),
            rt=get_float(__get_value(row, "Retention Time")),
            im_cv=get_float(__get_value(row, "Ion Mobility")),
            additional_information={"source": __serialize_pandas_series(row)},
        )
        csms.append(csm)
    return csms
